import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Async Greetings)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=10,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)

greeting = "Hello! How are you?"
lots_of_greetings = [greeting] * 50
num_of_greetings = len(lots_of_greetings)
num_said_greetings = 0
greeting1 = "Hello! How are you?"

# yields batch of results that are produced asynchronously and in parallel
for result in model.generate_async(
    lots_of_greetings,
    ordered=True,
    throw_on_error=True,
    hide_progressbar=False,
):
    num_said_greetings += 1
    print(f"[Progress {str(float(num_said_greetings/num_of_greetings)*100)}%]")
    print(f"\t {result.generated_text}")



import logging
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Union

from pydantic import BaseModel, Extra, root_validator

from genai.schemas.generate_params import GenerateParams

logger = logging.getLogger(__name__)


class StopReasonEnum(str, Enum):
    NOT_FINISHED = "Possibly more tokens to be streamed"
    MAX_TOKENS = "Maximum requested tokens reached"
    EOS_TOKEN = "End of sequence token encountered"
    CANCELLED = "Request canceled by the client"
    TIME_LIMIT = "Time limit reached"
    STOP_SEQUENCE = "Stop sequence encountered"
    TOKEN_LIMIT = "Token limit reached"
    ERROR = "Error encountered"


def alert_extra_fields_validator(cls, values) -> dict:
    """Reusable validator that logs a warning if we see undocmented fields returned by the API

    Args:
        cls (ModelMetaclass): the pydantic model that is being validated
        values (dict): The incoming values to be checkec

    Returns:
        dict: The values post validation
    """
    extra_fields = values.keys() - cls.__fields__.keys()

    if extra_fields:
        logger.debug(
            f"Extra fields missing from {cls.__name__}. Add Optional[Type] typing for these fields: {extra_fields}"
        )

    return values


class GenAiResponseModel(BaseModel, extra=Extra.allow):
    # Validators
    _extra_fields_warning = root_validator(allow_reuse=True)(alert_extra_fields_validator)


class TermsOfUseResult(GenAiResponseModel):
    tou_accepted: bool
    tou_accepted_at: datetime
    firstName: str
    lastName: str
    data_usage_consent: bool = False
    generate_default: dict = None


class TermsOfUse(GenAiResponseModel):
    results: TermsOfUseResult


class TextPosition(GenAiResponseModel):
    start: int
    end: int


class HAPResult(GenAiResponseModel):
    flagged: bool
    score: float
    success: bool
    position: TextPosition


class ModerationResult(GenAiResponseModel):
    hap: List[HAPResult]


class GeneratedToken(GenAiResponseModel):
    logprob: Optional[float]
    text: Optional[str]


class GenerateResult(GenAiResponseModel):
    generated_text: str
    generated_token_count: int
    input_token_count: Optional[int]
    stop_reason: str
    stop_sequence: Optional[str]
    generated_tokens: Optional[list[GeneratedToken]]
    input_text: Optional[str]
    seed: Optional[int]
    moderation: Optional[ModerationResult]


class GenerateResponse(GenAiResponseModel):
    id: str
    model_id: str
    created_at: str
    results: List[GenerateResult]


class GenerateStreamResponse(GenAiResponseModel):
    generated_text: Optional[str]
    generated_token_count: Optional[int]
    input_token_count: Optional[int]
    stop_reason: Optional[str]
    stop_sequence: Optional[str]
    generated_tokens: Optional[list[GeneratedToken]]
    input_text: Optional[str]
    seed: Optional[int]
    moderation: Optional[ModerationResult]


class TokenizeResult(GenAiResponseModel):
    token_count: Optional[int]
    tokens: Optional[List[str]]
    input_text: Optional[str]


class TokenizeResponse(GenAiResponseModel):
    model_id: str
    created_at: str
    results: List[TokenizeResult]


class HistoryResultRequest(GenAiResponseModel):
    inputs: list[str]
    model_id: str
    parameters: GenerateParams


class HistoryResult(GenAiResponseModel):
    id: str
    duration: int
    request: HistoryResultRequest
    status: str
    created_at: datetime
    response: Union[GenerateResponse, Any]


class HistoryResponse(GenAiResponseModel):
    results: List[HistoryResult]
    totalCount: int


class ErrorExtensionStateParam(GenAiResponseModel):
    comparison: str
    limit: int


class ErrorExtensionState(GenAiResponseModel):
    instancePath: str
    params: ErrorExtensionStateParam


class ErrorExtensions(GenAiResponseModel):
    code: str
    state: list[ErrorExtensionState]


class ErrorResponse(GenAiResponseModel):
    status_code: int
    error: str
    message: str
    extensions: Optional[ErrorExtensions]


class WatsonxTemplate(GenAiResponseModel):
    id: str
    name: str
    value: str
    created_at: datetime
    data: Optional[dict]


class WatsonxTemplatesResponse(GenAiResponseModel):
    results: list[WatsonxTemplate]
    totalCount: int


class WatsonxRenderedPrompts(GenAiResponseModel):
    results: list[str]


class FileFormatResult(GenAiResponseModel):
    id: int
    name: str


class FileInfoResult(GenAiResponseModel):
    id: str
    bytes: str
    file_name: str
    purpose: str
    storage_provider_location: Optional[str]
    created_at: datetime
    file_formats: List[FileFormatResult]


class FilesListResponse(GenAiResponseModel):
    results: List[FileInfoResult]
    totalCount: int


class TuneParameters(GenAiResponseModel):
    accumulate_steps: Optional[int]
    batch_size: Optional[int]
    learning_rate: Optional[float]
    max_input_tokens: Optional[int]
    max_output_tokens: Optional[int]
    num_epochs: Optional[int]
    num_virtual_tokens: Optional[int]
    verbalizer: Optional[str]


class TuneInfoResult(GenAiResponseModel):
    id: str
    name: str
    model_id: str
    model_name: str
    method_id: Optional[str]
    method_name: Optional[str]
    status: str
    task_id: str
    task_name: Optional[str]
    parameters: Optional[TuneParameters]
    created_at: datetime
    preferred: Optional[bool]
    datapoints: Optional[dict]
    validation_files: Optional[list]
    training_files: Optional[list]
    evaluation_files: Optional[list]
    status_message: Optional[str]
    started_at: Optional[datetime]


class TunesListResponse(GenAiResponseModel):
    results: List[TuneInfoResult]
    totalCount: int


class TrainingFilesParameters(GenAiResponseModel):
    id: str
    file_name: str
    created_at: str


class TuneGetResponse(GenAiResponseModel):
    results: Optional[TuneInfoResult]


class TuneMethodsInfo(GenAiResponseModel):
    id: str
    name: str


class TuneMethodsGetResponse(GenAiResponseModel):
    results: Optional[List[TuneMethodsInfo]]


class ModelCard(GenAiResponseModel):
    id: Optional[str]
    name: Optional[str]
    size: Optional[str]
    source_model_id: Optional[str]
    token_limit: Optional[Union[int, Any]]


class ModelList(GenAiResponseModel):
    results: list[ModelCard]


class GenerateLimits(GenAiResponseModel):
    tokenCapacity: int
    tokensUsed: int



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.extensions.huggingface.agent import IBMGenAIAgent
from genai.schemas import GenerateParams

load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

creds = Credentials(api_key, api_endpoint)
params = GenerateParams(min_new_tokens=10, max_new_tokens=200)

agent = IBMGenAIAgent(credentials=creds, model="meta-llama/llama-2-70b-chat", params=params)

agent.chat("Download the text from the given url", url="https://research.ibm.com/blog/analog-ai-chip-low-power")
agent.chat("Summarize the downloaded text")



from genai.exceptions import GenAiException

__all__ = ["to_genai_error"]


def to_genai_error(e: Exception) -> GenAiException:
    if isinstance(e, GenAiException):
        return e

    return GenAiException(e)



import unittest

import pytest

from genai.schemas import GenerateParams, ReturnOptions
from genai.utils.request_utils import sanitize_params

# API Reference : https://workbench.res.ibm.com/docs/api-reference#generate

SERVICE_URL = "https://workbench-api.res.ibm.com"
KEY = "pak-"


@pytest.mark.unit
class TestServiceUtils(unittest.TestCase):
    def test_sanitize_params_with_params(self):
        input_genParams = GenerateParams(decoding_method="greedy", return_options=ReturnOptions(input_text=True))

        expected_genParams = {"decoding_method": "greedy", "return_options": {"input_text": True}}

        sanitized_dict = sanitize_params(input_genParams)
        self.assertEqual(expected_genParams, sanitized_dict)



from unittest.mock import MagicMock, patch

import pytest

from genai.credentials import Credentials
from genai.exceptions import GenAiException
from genai.schemas.responses import WatsonxTemplate, WatsonxTemplatesResponse
from genai.services import PromptTemplateManager
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestPromptTemplateManager:
    def setup_method(self):
        self.credentials = Credentials("KEY", "ENDPOINT")
        self.string_template = """
            {{ instruction }}
            {{#examples}}

            {{input}}
            {{output}}

            {{/examples}}
            {{input}}
        """
        self.name = "My template"

        self.expected_resp = SimpleResponse.prompt_template(template=self.string_template, name=self.name)

        self.template = WatsonxTemplate.parse_obj(self.expected_resp["results"])
        self.all_templates = WatsonxTemplatesResponse.parse_obj({"results": [self.template], "totalCount": 1})

    @patch("genai.services.RequestHandler.post")
    def test_save_template(self, mocked_post_request):
        expected = MagicMock(status_code=200)
        expected.json.return_value = self.expected_resp
        mocked_post_request.return_value = expected

        template = PromptTemplateManager.save_template(
            credentials=self.credentials, template=self.string_template, name=self.name
        )

        assert isinstance(template, WatsonxTemplate)
        assert template.value == self.string_template
        assert template.name == self.name

    @patch("genai.services.RequestHandler.post", side_effect=Exception("Ooops"))
    def test_save_template_exception(self, mocked_post_request):
        expected = MagicMock(status_code=200)
        expected.json.return_value = self.expected_resp
        mocked_post_request.return_value = expected

        with pytest.raises(GenAiException, match="Ooops"):
            PromptTemplateManager.save_template(
                credentials=self.credentials, template=self.string_template, name=self.name
            )

    @patch("genai.services.RequestHandler.put")
    def test_update_template(self, mocked_post_request):
        expected = MagicMock(status_code=200)
        expected.json.return_value = self.expected_resp
        mocked_post_request.return_value = expected

        _id = self.template.id
        template = PromptTemplateManager.update_template(
            credentials=self.credentials, template=self.string_template, name=self.name, id=_id
        )

        assert isinstance(template, WatsonxTemplate)
        assert template.value == self.string_template
        assert template.name == self.name

    @patch("genai.services.RequestHandler.put", side_effect=Exception("Ooops"))
    def test_update_template_exception(self, mocked_post_request):
        expected = MagicMock(status_code=200)
        expected.json.return_value = self.expected_resp
        mocked_post_request.return_value = expected

        with pytest.raises(GenAiException, match="Ooops"):
            _id = self.template.id
            PromptTemplateManager.update_template(
                credentials=self.credentials, template=self.string_template, name=self.name, id=_id
            )

    @patch("genai.services.PromptTemplateManager.load_template_by_name")
    @patch("genai.services.PromptTemplateManager.load_template_by_id")
    def test_load_template(self, mock_load_by_id, mock_load_by_name):
        _id = "my_super_id"
        PromptTemplateManager.load_template(credentials=self.credentials, id=_id)
        mock_load_by_id.assert_called_with(credentials=self.credentials, id=_id)

        _name = "my_super_name"
        PromptTemplateManager.load_template(credentials=self.credentials, name=_name)
        mock_load_by_name.assert_called_with(credentials=self.credentials, name=_name)

        error = (
            "Provide either name or id of prompt to be fetch."
            + "\nIf you want to list all templates, use the load_all_templates method."
        )
        with pytest.raises(GenAiException, match=error):
            PromptTemplateManager.load_template(credentials=self.credentials)

    @patch("genai.services.RequestHandler.get")
    def test_load_all_template(self, mock_get):
        expected = MagicMock(status_code=200)
        expected.json.return_value = {"results": [self.template.dict()], "totalCount": 1}
        mock_get.return_value = expected

        resp = PromptTemplateManager.load_all_templates(credentials=self.credentials)

        assert isinstance(resp, WatsonxTemplatesResponse)
        assert isinstance(resp.results[0], WatsonxTemplate)
        assert resp.totalCount == 1

    @patch("genai.services.RequestHandler.get")
    def test_load_all_template_404(self, mock_get):
        expected = MagicMock(status_code=404)
        mock_get.return_value = expected

        with pytest.raises(GenAiException):
            PromptTemplateManager.load_all_templates(credentials=self.credentials)

    @patch("genai.services.RequestHandler.get")
    def test_load_template_by_id(self, mock_get):
        expected = MagicMock(status_code=200)
        expected.json.return_value = self.expected_resp
        mock_get.return_value = expected

        _id = self.template.id
        template = PromptTemplateManager.load_template_by_id(credentials=self.credentials, id=_id)

        assert isinstance(template, WatsonxTemplate)
        assert template.value == self.string_template
        assert template.name == self.name

    @patch("genai.services.RequestHandler.get")
    def test_load_template_by_id_not_found(self, mock_get):
        expected = MagicMock(status_code=404)
        expected.json.return_value = self.expected_resp
        mock_get.return_value = expected

        with pytest.raises(Exception):
            _id = self.template.id
            PromptTemplateManager.load_template_by_id(credentials=self.credentials, id=_id)

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.get")
    def test_load_template_by_name(self, mock_get, mock_load_all_templates):
        expected = MagicMock(status_code=200)
        mock_get.return_value = expected
        mock_load_all_templates.return_value = self.all_templates

        name = self.template.name
        template = PromptTemplateManager.load_template_by_name(credentials=self.credentials, name=name)

        assert isinstance(template, WatsonxTemplate)
        assert template.id == self.template.id
        assert template.name == self.template.name

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.get")
    def test_load_template_by_name_not_found(self, mock_get, mock_load_all_templates):
        no_templates = WatsonxTemplatesResponse.parse_obj({"results": [], "totalCount": 0})

        expected = MagicMock(status_code=200)
        mock_get.return_value = expected
        mock_load_all_templates.return_value = no_templates

        name = self.template.name
        with pytest.raises(GenAiException, match=f"No template found with name {name}"):
            PromptTemplateManager.load_template_by_name(credentials=self.credentials, name=name)

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.get")
    def test_load_template_by_name_too_many_found(self, mock_get, mock_load_all_templates):
        no_templates = WatsonxTemplatesResponse.parse_obj({"results": [self.template, self.template], "totalCount": 2})

        expected = MagicMock(status_code=200)
        mock_get.return_value = expected
        mock_load_all_templates.return_value = no_templates

        name = self.template.name
        with pytest.raises(Exception, match=f"More than one template found with name {name}"):
            PromptTemplateManager.load_template_by_name(credentials=self.credentials, name=name)

    @patch("genai.services.PromptTemplateManager.delete_template_by_name")
    @patch("genai.services.PromptTemplateManager.delete_template_by_id")
    def test_delete_template(self, mock_delete_by_id, mock_delete_by_name):
        _id = "my_super_id"
        PromptTemplateManager.delete_template(credentials=self.credentials, id=_id)
        mock_delete_by_id.assert_called_with(credentials=self.credentials, id=_id)

        _name = "my_super_name"
        PromptTemplateManager.delete_template(credentials=self.credentials, name=_name)
        mock_delete_by_name.assert_called_with(credentials=self.credentials, name=_name)

        error = "Provide either name or id of prompt to be deleted."
        with pytest.raises(GenAiException, match=error):
            PromptTemplateManager.delete_template(credentials=self.credentials)

    @patch("genai.services.RequestHandler.delete")
    def test_delete_template_by_id(self, mock_delete):
        expected = MagicMock(status_code=204)
        mock_delete.return_value = expected

        _id = "prompt_template_id"
        template = PromptTemplateManager.delete_template_by_id(credentials=self.credentials, id=_id)

        assert isinstance(template, str)
        assert template == _id

    @patch("genai.services.RequestHandler.delete")
    def test_delete_template_by_id_not_found(self, mock_delete):
        expected = MagicMock(status_code=404)
        mock_delete.return_value = expected

        with pytest.raises(Exception):
            _id = "prompt_template_id"
            PromptTemplateManager.delete_template_by_id(credentials=self.credentials, id=_id)

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.delete")
    def test_delete_template_by_name(self, mock_delete, mock_load_all_templates):
        expected = MagicMock(status_code=204)
        mock_delete.return_value = expected
        mock_load_all_templates.return_value = self.all_templates

        _id = self.template.id
        name = self.template.name

        template = PromptTemplateManager.delete_template_by_name(credentials=self.credentials, name=name)

        assert isinstance(template, str)
        assert template == _id

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.delete")
    def test_delete_template_by_name_not_found(self, mock_delete, mock_load_all_templates):
        no_templates = WatsonxTemplatesResponse.parse_obj({"results": [], "totalCount": 0})

        expected = MagicMock(status_code=204)
        mock_delete.return_value = expected
        mock_load_all_templates.return_value = no_templates

        name = "not the template you are looking for"
        with pytest.raises(Exception, match=f"No template found for name {name}"):
            PromptTemplateManager.delete_template_by_name(credentials=self.credentials, name=name)

    @patch("genai.services.PromptTemplateManager.load_all_templates")
    @patch("genai.services.RequestHandler.delete")
    def test_delete_template_by_name_too_many_found(self, mock_delete, mock_load_all_templates):
        all_templates = WatsonxTemplatesResponse.parse_obj({"results": [self.template, self.template], "totalCount": 2})

        expected = MagicMock(status_code=204)
        mock_delete.return_value = expected
        mock_load_all_templates.return_value = all_templates

        name = self.template.name
        with pytest.raises(Exception, match=f"More than one template found for name {name}"):
            PromptTemplateManager.delete_template_by_name(credentials=self.credentials, name=name)

    @patch("genai.services.RequestHandler.post")
    def test_render_template(self, mock_post):
        expected = MagicMock(status_code=200)
        expected.json.return_value = {"results": ["rendered output 1", "rendered output 2"]}
        mock_post.return_value = expected

        results = PromptTemplateManager.render_watsonx_prompts(
            credentials=self.credentials, inputs=["some inputs"], data={}
        )

        assert len(results) == 2
        assert isinstance(results[0], str)

    @patch("genai.services.RequestHandler.post")
    def test_render_template_exception(self, mock_post):
        mock_post.return_value = MagicMock(status_code=404)

        with pytest.raises(GenAiException):
            PromptTemplateManager.render_watsonx_prompts(credentials=self.credentials, inputs=["some inputs"], data={})



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.prompt_pattern import PromptPattern
from genai.utils.search_space_params import grid_search_generate_params

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint> (optional) DEFAULT_API = "https://workbench-api.res.ibm.com"
load_dotenv()
API_KEY = os.getenv("GENAI_KEY", None)
API_ENDPOINT = os.getenv("GENAI_API", None)


print("\n------------- Example (String Replacement)-------------\n")

# use the dictionary to define the search space, keep the keys as
# the same from GenerateParams and use a list for the values to search
my_space_params = {
    "decoding_method": ["sample"],
    "max_new_tokens": [10, 20],
    "min_new_tokens": [1, 2],
    "temperature": [0.7, 0.8, 0.9, 1.5],
}

creds = Credentials(api_key=API_KEY, api_endpoint=API_ENDPOINT)
pt = PromptPattern.from_str("The capital of {{country}} is {{capital}}. The capital of Canada is")
pt.sub("capital", "Madrid").sub("country", "Spain")

# generate all combinations of parameters, returns a list of GenerateParams
generate_params_list = grid_search_generate_params(my_space_params)

for params in generate_params_list:
    model = Model("google/flan-t5-xxl", params=params, credentials=creds)
    responses = model.generate_async([str(pt)])

    print(f"Used params: \n{params} \n")
    print(pt)
    for response in responses:
        print(f"Generated text: {response.generated_text}")
    print("------------------------------------")



from unittest.mock import MagicMock, patch

import pytest

from genai.schemas import GenerateParams, ReturnOptions
from genai.services import ServiceInterface
from genai.utils.request_utils import sanitize_params
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestServiceInterface:
    def setup_method(self):
        # mock object for the API call
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def params(seld):
        return GenerateParams(decoding_method="greedy", return_options=ReturnOptions(input_text=True))

    @patch("genai.services.RequestHandler.patch")
    def test_tou(self, mocked_post_request):
        expected_resp = SimpleResponse.terms_of_use()
        expected = MagicMock(status_code=200, json=expected_resp)
        mocked_post_request.return_value = expected

        resp = self.service.terms_of_use(True)

        assert resp == expected
        assert resp.status_code == 200

    @patch(
        "genai.services.RequestHandler.patch",
        side_effect=Exception("some general error"),
    )
    def test_tou_exception(self, mock):
        with pytest.raises(BaseException, match="some general error"):
            self.service.terms_of_use(True)

    def test_sanitize_params(self, params):
        new_dict = sanitize_params(params=params)

        assert isinstance(new_dict, dict)
        assert "min_new_tokens" not in new_dict



from unittest.mock import MagicMock, patch

import pytest

from genai import Credentials, Metadata
from genai.schemas.responses import HistoryResponse, TermsOfUse
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestMetadata:
    @patch("genai.services.ServiceInterface.terms_of_use")
    def test_accept_TOU(self, mock_requests):
        """Tests that we can accept the TOU"""

        TOU_RESPONSE = SimpleResponse.terms_of_use()

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = TOU_RESPONSE
        mock_requests.return_value = mock_response

        # Build up our Model Object
        creds = Credentials("TEST_API_KEY")

        # Instantiate the GENAI Proxy Object
        model_meta = Metadata(creds)

        tou_response = model_meta.accept_terms_of_use()
        original_tou = TermsOfUse(**TOU_RESPONSE)
        assert tou_response == original_tou

    @patch("genai.services.ServiceInterface.history")
    def test_history(self, mock_requests):
        """Tests that we can get the History"""

        HISTORY_RESPONSE = SimpleResponse.history()

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = HISTORY_RESPONSE
        mock_requests.return_value = mock_response

        # Build up our Model Object
        creds = Credentials("TEST_API_KEY")

        # Instantiate the GENAI Proxy Object
        model_meta = Metadata(creds)

        history_response = model_meta.get_history()
        original_history_response = HistoryResponse(**HISTORY_RESPONSE)
        assert history_response == original_history_response



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.metadata import Metadata
from genai.schemas.history_params import HistoryParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (History)-------------\n")

creds = Credentials(api_key, api_endpoint)
object = Metadata(creds)

params = HistoryParams(
    limit=8,
    offset=0,
    status="SUCCESS",
    origin="API",
)

history_response = object.get_history(params)
print(f"Response: \n{history_response}\n")



import httpx

__all__ = ["HttpProvider"]


class HttpProvider:
    default_http_client_options: dict = {}
    default_http_transport_options: dict = {}

    @classmethod
    def get_client(cls, **kwargs):
        options = {**cls.default_http_client_options, **kwargs}
        return httpx.Client(**options)

    @classmethod
    def get_async_client(cls, **kwargs):
        options = {**cls.default_http_client_options, **kwargs}
        return httpx.AsyncClient(**options)

    @classmethod
    def get_async_transport(cls, **kwargs):
        options = {**cls.default_http_transport_options, **kwargs}
        return httpx.AsyncHTTPTransport(**options)



from unittest.mock import MagicMock, patch

import pytest

from genai.schemas import HistoryParams
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse

# API Reference : https://workbench.res.ibm.com/docs/api-reference#generate


@pytest.mark.unit
class TestHistory:
    def setup_method(self):
        # mock object for the API call
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        # test all GenerateParams fields

    @pytest.fixture
    def params(self):
        return HistoryParams(
            limit=10,
            offset=0,
            status="SUCCESS",
            origin="API",
        )

    @patch("genai.services.RequestHandler.get")
    def test_history_api_call(self, mock_requests, params):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.history()
        mock_requests.return_value = mock_response

        g = self.service.history(params=params)
        assert g == mock_response

    @patch("genai.services.RequestHandler.get", side_effect=Exception("some general error"))
    def test_history_general_exception(self, mock):
        with pytest.raises(BaseException, match="some general error"):
            self.service.history()



from unittest.mock import MagicMock, patch

import pytest

from genai.credentials import Credentials
from genai.prompt_pattern import PromptPattern
from genai.schemas.responses import WatsonxTemplate
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestPromptPattern:
    def setup_method(self):
        self.credentials = Credentials("KEY", "ENDPOINT")
        self.string_template = "{{input}}:{{ouptu}}"
        self.name = "io"

        self.expected_resp = SimpleResponse.prompt_template(template=self.string_template, name=self.name)
        self.template = WatsonxTemplate.parse_obj(self.expected_resp["results"])

    @patch("genai.services.PromptTemplateManager.save_template")
    def test_from_watsonx(self, save):
        save.return_value = self.template

        pt = PromptPattern.from_watsonx(credentials=self.credentials, template=self.string_template, name=self.name)
        assert isinstance(pt, PromptPattern)
        assert str(pt) == self.string_template
        assert pt.watsonx == self.template

    @patch("genai.services.PromptTemplateManager.load_template")
    @patch("genai.services.PromptTemplateManager.update_template")
    @patch("genai.services.PromptTemplateManager.save_template")
    def test_from_watsonx_logic(self, save, update, load):
        # Cases :
        # fetching an existing template : name OR id
        # updating an existing template : template + (name OR id)
        # creating a new template       : template + name

        # Case 1 : fetching an existing template
        PromptPattern.from_watsonx(credentials=self.credentials, name=self.name)
        load.assert_called_with(credentials=self.credentials, name=self.name, id=None)

        PromptPattern.from_watsonx(credentials=self.credentials, id="_id")
        load.assert_called_with(credentials=self.credentials, id="_id", name=None)

        # Case 2 : updating an existing template
        load.return_value = self.template
        _id = self.template.id
        _name = self.template.name

        PromptPattern.from_watsonx(credentials=self.credentials, id=_id, template="Instruction: {{instruction}}")
        update.assert_called_with(
            credentials=self.credentials, id=_id, template="Instruction: {{instruction}}", name=_name
        )

        PromptPattern.from_watsonx(credentials=self.credentials, name=_name, template="Instruction: {{instruction}}")
        update.assert_called_with(
            credentials=self.credentials, id=_id, template="Instruction: {{instruction}}", name=_name
        )

        # Case 3 : saving a new template
        load.side_effect = Exception(MagicMock(status=404), "not found")
        _name = self.template.name

        PromptPattern.from_watsonx(credentials=self.credentials, template="Instruction: {{instruction}}", name=_name)
        save.assert_called_with(credentials=self.credentials, template="Instruction: {{instruction}}", name=_name)

    @patch("genai.services.RequestHandler.delete")
    @patch("genai.services.PromptTemplateManager.save_template")
    def test_pp_watsonx_delete(self, save, delete):
        save.return_value = self.template
        delete.return_value = MagicMock(status_code=204)

        pt = PromptPattern.from_watsonx(credentials=self.credentials, template=self.string_template, name=self.name)

        id = pt.delete()
        assert id == self.template.id



import os
import pathlib

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
PATH = pathlib.Path(__file__).parent.resolve()

print("\n------------- Example (PromptPatterns) -------------\n")

params = GenerateParams(
    decoding_method="greedy",
    max_new_tokens=20,
    min_new_tokens=1,
    stream=False,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)

# (1) Prompt
prompt = "Is McDonald's or Burger King better?"

print("INPUT>> " + prompt + "\n")
responses = model.generate([prompt])
gen_response = responses[0].generated_text.strip()

# (2) Internal monolog
print("[Internal monologue]")
print(f"[My answer]: {gen_response}")

# (2.1) critic myself with template
pt = PromptPattern.from_file(str(PATH) + os.sep + "templates" + os.sep + "self-reflection.yaml")
pt.sub("prompt", prompt)
pt.sub("response", gen_response)
print(f"[Self reflection]: {pt}")

# (2.2) critic my answer
# adjust params
params.min_new_tokens = 1
params.max_new_tokens = 1
responses = model.generate([pt])
gen_control_ans = responses[0].generated_text
print(f"[Self reflection answer]: {gen_control_ans}" + "\n")

# (3) self-control flow
if "yes" in gen_control_ans.lower():
    print("<< Sorry I should not comment.")
else:
    print("<<OUTPUT " + gen_response)



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Greetings Tokens)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=10,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)

greeting1 = "Hello! How are you?"
greeting2 = "I am fine and you?"

# just get the token counts
responses = model.tokenize_as_completed([greeting1, greeting2])
for response in responses:
    print(f"Generated text: {response}")

# get the token counts and tokens
responses = model.tokenize_as_completed([greeting1, greeting2], return_tokens=True)
for response in responses:
    print(f"Generated text: {response}")



#
# NOTE: This example has been written to enable an end-user
# to quickly try prompt-tuning. In order to obtain better
# performance, a user would need to experiment with the
# number of observations and tuning hyperparameters
#

import os
import time
from pathlib import Path

try:
    from datasets import load_dataset
except ImportError:
    print("Install datasets: it is a pre-requisite to run this example")

try:
    import pandas as pd
except ImportError:
    print("Install pandas: it is a pre-requisite to run this example")

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas.generate_params import GenerateParams
from genai.schemas.tunes_params import CreateTuneHyperParams, TunesListParams
from genai.services import FileManager
from genai.services.tune_manager import TuneManager

load_dotenv()
num_training_samples = 100
num_validation_samples = 20
data_root = Path(__file__).parent.resolve() / "data"
training_file = data_root / "fpb_train.jsonl"
validation_file = data_root / "fpb_validation.jsonl"


def create_dataset():
    Path(data_root).mkdir(parents=True, exist_ok=True)
    if training_file.exists() and validation_file.exists():
        return
    data = load_dataset(
        "financial_phrasebank",
        "sentences_allagree",
    )
    df = pd.DataFrame(data["train"]).sample(n=num_training_samples + num_validation_samples)
    df.rename(columns={"sentence": "input", "label": "output"}, inplace=True)
    df["output"] = df["output"].astype(str)
    train_jsonl = df.iloc[:num_training_samples].to_json(orient="records", lines=True, force_ascii=True)
    validation_jsonl = df.iloc[num_training_samples:].to_json(orient="records", lines=True, force_ascii=True)
    with open(training_file, "w") as fout:
        fout.write(train_jsonl)
    with open(validation_file, "w") as fout:
        fout.write(validation_jsonl)


def upload_files(creds, update=True):
    fileinfos = FileManager.list_files(credentials=creds).results
    filenames_to_id = {f.file_name: f.id for f in fileinfos}
    for filepath in [training_file, validation_file]:
        filename = filepath.name
        if update and filename in filenames_to_id:
            print(f"File already present: Overwriting {filename}")
            FileManager.delete_file(credentials=creds, file_id=filenames_to_id[filename])
            FileManager.upload_file(credentials=creds, file_path=str(filepath), purpose="tune")
        if filename not in filenames_to_id:
            print(f"File not present: Uploading {filename}")
            FileManager.upload_file(credentials=creds, file_path=str(filepath), purpose="tune")


def get_file_ids(creds):
    fileinfos = FileManager.list_files(credentials=creds).results
    training_file_ids = [f.id for f in fileinfos if f.file_name == training_file.name]
    validation_file_ids = [f.id for f in fileinfos if f.file_name == validation_file.name]
    return training_file_ids, validation_file_ids


def get_creds():
    api_key = os.getenv("GENAI_KEY", None)
    endpoint = os.getenv("GENAI_API", None)
    creds = Credentials(api_key=api_key, api_endpoint=endpoint)
    return creds


if __name__ == "__main__":
    creds = get_creds()
    create_dataset()
    upload_files(creds, update=True)

    model = Model("google/flan-t5-xl", params=None, credentials=creds)
    hyperparams = CreateTuneHyperParams(num_epochs=2, verbalizer='classify { "0", "1", "2" } Input: {{input}} Output:')
    training_file_ids, validation_file_ids = get_file_ids(creds)

    tuned_model = model.tune(
        name="classification-mpt-tune-api",
        method="mpt",
        task="classification",
        hyperparameters=hyperparams,
        training_file_ids=training_file_ids,
        validation_file_ids=validation_file_ids,
    )

    status = tuned_model.status()
    while status not in ["FAILED", "HALTED", "COMPLETED"]:
        print(status)
        time.sleep(20)
        status = tuned_model.status()
    else:
        if status in ["FAILED", "HALTED"]:
            print("Model tuning failed or halted")
        else:
            print("Model info:\n")
            print(tuned_model.info())
            time.sleep(5)

            prompt = input("Enter a prompt:\n")
            genparams = GenerateParams(
                decoding_method="greedy",
                max_new_tokens=50,
                min_new_tokens=1,
            )
            print("Answer = ", tuned_model.generate([prompt])[0].generated_text)
            time.sleep(5)

            print("~~~~~~~ List of all models ~~~~~~")
            for m in Model.models(credentials=creds):
                print(m, "\n")
            time.sleep(10)

            print("~~~~~~~ Getting list of all tuned models with TuneManager ~~~~~")

            list_params = TunesListParams(limit=50, offset=0)

            tune_list = TuneManager.list_tunes(credentials=creds, params=list_params)
            print("\n\nList of tunes: \n\n")
            for tune in tune_list.results:
                print(tune, "\n")
            time.sleep(10)

            tune_get_result = TuneManager.get_tune(credentials=creds, tune_id=tuned_model.model)
            print(
                "\n\n~~~~~ Metadata for a single tune with TuneManager ~~~~: \n\n",
                tune_get_result,
            )
            time.sleep(5)

            print("~~~~~~~ Downloading tuned model assets ~~~~~")
            to_download_assets = input("Download tuned model assets? (y/N):\n")
            if to_download_assets == "y":
                tuned_model.download()

            time.sleep(5)

            print("~~~~~~~ Deleting a tuned model ~~~~~")
            to_delete = input("Delete this model? (y/N):\n")
            if to_delete == "y":
                tuned_model.delete()



from unittest.mock import MagicMock, patch

import pytest

from genai.schemas import GenerateParams
from genai.services import RequestHandler, ServiceInterface


@pytest.mark.unit
class TestRequestHandler:
    def setup_method(self):
        self.model = "somemodel"
        self.inputs = ["a"]

    @pytest.fixture
    def params(seld):
        return GenerateParams(decoding_method="greedy").dict(by_alias=True, exclude_none=True)

    def test_metadata_post(self, params):
        headers, json_data, _ = RequestHandler._metadata(
            method="POST",
            key="API_KEY",
            model_id=self.model,
            inputs=self.inputs,
            parameters=params,
        )

        assert "Content-Type" in headers
        assert headers["Content-Type"] == "application/json"

        assert "Authorization" in headers
        assert headers["Authorization"] == "Bearer API_KEY"

        assert json_data == {
            "model_id": self.model,
            "inputs": self.inputs,
            "parameters": {"decoding_method": "greedy"},
        }

    def test_metadata_get(self):
        headers, _, _ = RequestHandler._metadata(method="GET", key="API_KEY")

        assert "Authorization" in headers
        assert headers["Authorization"] == "Bearer API_KEY"

    @patch("httpx.Client.get")
    def test_get(self, mock: MagicMock):
        expected_resp = {"some": "history"}
        mock.return_value = expected_resp

        s = ServiceInterface(service_url="SERVICE_URL", api_key="KEY")
        his = s.history(params={"limit": 1, "status": "SUCCESS"})

        assert his == expected_resp

    @patch("httpx.Client.post")
    def test_post(self, mock: MagicMock):
        expected_resp = "some tokens"
        mock.return_value = expected_resp

        s = ServiceInterface(service_url="SERVICE_URL", api_key="KEY")
        toekn = s.tokenize(model="model", inputs=["input"])

        assert toekn == expected_resp



import asyncio
import logging
import os
import random

import httpx
from dotenv import load_dotenv

from genai.model import Credentials, Model
from genai.schemas import GenerateParams, TokenParams
from genai.services.connection_manager import ConnectionManager
from genai.services.request_handler import RequestHandler

logging.basicConfig(level="ERROR")


class FlakyRequestHandler(RequestHandler):
    @staticmethod
    async def flaky_async_generate(
        endpoint: str, key: str, model_id: str = None, inputs: list = None, parameters: dict = None, options=None
    ):
        """Low level API for async /generate request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, json_data, _ = RequestHandler._metadata(
            method="POST",
            key=key,
            model_id=model_id,
            inputs=inputs,
            parameters=parameters,
            options=options,
        )

        response = None
        totalsleep = 0
        for attempt in range(0, ConnectionManager.MAX_RETRIES_GENERATE):
            response = await ConnectionManager.async_generate_client.post(endpoint, headers=headers, json=json_data)
            choice = random.randint(0, 2)
            if choice == 0:
                response.status_code = httpx.codes.SERVICE_UNAVAILABLE
            elif choice == 1:
                response.status_code = httpx.codes.TOO_MANY_REQUESTS
            else:
                pass
            if response.status_code in [httpx.codes.SERVICE_UNAVAILABLE, httpx.codes.TOO_MANY_REQUESTS]:
                totalsleep += 2 ** (attempt + 1)
                print(
                    "Prompt = {}, Choice = {}, Current sleep {} seconds, Total sleep = {}".format(
                        json_data["inputs"], choice, 2 ** (attempt + 1), totalsleep
                    )
                )
                await asyncio.sleep(2 ** (attempt + 1))
            else:
                break
        return response


RequestHandler.async_generate = FlakyRequestHandler.flaky_async_generate

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_url = os.getenv("GENAI_API", None)
creds = Credentials(api_key=api_key, api_endpoint=api_url)  # credentials object to access GENAI


# Instantiate parameters for text generation
generate_params = GenerateParams(decoding_method="sample", max_new_tokens=5, min_new_tokens=1)
tokenize_params = TokenParams(return_tokens=True)


flan_ul2 = Model("google/flan-ul2", params=generate_params, credentials=creds)
prompts = ["Generate a random number > {}: ".format(i) for i in range(25)]
for response in flan_ul2.generate_async(prompts, ordered=True):
    pass



from typing import Any, Type, TypeVar

T = TypeVar("T")


def to_model_instance(params: Any, Model: Type[T]) -> T:
    if params is None:
        return Model()

    if isinstance(params, Model):
        return params.copy()

    if isinstance(params, dict):
        return Model(**params)

    raise ValueError(f"The 'params' property should be an instance of {Model.__name__} class. Current value: {params}")



import os
import pathlib

try:
    import pandas as pd
except ImportError:
    raise ImportError("Could not import pandas: Please install ibm-generative-ai[pandas] extension.")

from dotenv import load_dotenv

import genai.extensions.pandas  # noqa: F401
from genai.credentials import Credentials
from genai.model import Model
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams

#
# In this demo, the following dataset was used:
#
# Gorman KB, Williams TD, Fraser WR (2014) "Ecological Sexual Dimorphism and Environmental Variability within a Community of Antarctic Penguins (Genus Pygoscelis)." PLoS ONE 9(3): e90081. doi:10.1371/journal.pone.0090081 # noqa

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
PATH = pathlib.Path(__file__).parent.resolve()

print("\n------------- Example (Pandas Dataframe Replacement)-------------\n")

params = GenerateParams(
    decoding_method="greedy",
    max_new_tokens=15,
    min_new_tokens=1,
    stream=False,
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)


csv_path = str(PATH) + os.sep + "assets" + os.sep + "penguins.csv"
df = pd.read_csv(csv_path, index_col=0)

pt = PromptPattern.from_str("{{species}}: {{island}}, {{flipper_length_mm}}, {{year}}\n")
list_of_prompts = pt.pandas.sub_from_dataframe(dataframe=df, start_index=15, n=2, strategy="sample")

print("-----------------------")
print("generated prompt")
print(list_of_prompts)
print("Number of prompts: ", len(list_of_prompts))
print("-----------------------")

responses = model.generate_as_completed(list_of_prompts)
for response in responses:
    print(f"Generated text: {response.generated_text}")



# flake8: noqa


class Descriptions:
    # LengthPenalty
    DECAY_FACTOR = "Represents the factor of exponential decay and must be > 1.0. Larger values correspond to more aggressive decay."
    START_INDEX = "A number of generated tokens after which this should take effect."

    # Return
    RETURN = "key-value pairs."
    INPUT_TEXT = "Include input text."
    GENERATED_TOKEN = "Include list of individual generated tokens. 'Extra' token information is included based on the other flags below."
    INPUT_TOKEN = "Include list of input tokens. 'Extra' token information is included based on the other flags here, but only for decoder-only models."
    TOKEN_LOGPROBS = "Include logprob for each returned token. Applicable only if generated_tokens == true and/or input_tokens == true."
    TOKEN_RANKS = (
        "Include rank of each returned token. Applicable only if generated_tokens == true and/or input_tokens == true."
    )
    TOP_N_TOKENS = "Include top n candidate tokens at the position of each returned token. The maximum value permitted is 5, but more may be returned if there is a tie for nth place. Applicable only if generated_tokens == true and/or input_tokens == true."

    # Params.Generate
    DECODING_METHOD = (
        "Decoding method used for generation. Options are greedy or sample. Defaults to sample if not specified."
    )
    LENGTH_PENALTY = "It can be used to exponentially increase the likelihood of the text generation terminating once a specified number of tokens have been generated."
    MAX_NEW_TOKEN = "The maximum number of new tokens to be generated. The range is 1 to 1024, defaults to 20."
    MIN_NEW_TOKEN = "If stop sequences are given, they are ignored until minimum tokens are generated."
    RANDOM_SEED = "Manually passed seed to initialize the randomization for experimental repeatability. The range is 1 to 9999. Valid only with decoding_method=sample."
    STOP_SEQUENCES = "Stop sequences are one or more strings which will cause the text generation to stop if/when they are produced as part of the output. Stop sequences encountered prior to the minimum number of tokens being generated will be ignored."
    STREAM = "Enables to stream partial progress as server-sent events. Defaults to false."
    TEMPERATURE = "The value used to module the next token probabilities. The range is 0.05 to 2.00, a value set to 0.05 would make it deterministic. Valid only with decoding_method=sample."
    TIME_LIMIT = "Time limit in milliseconds - if not completed within this time, generation will stop. The text generated so far will be returned along with the TIME_LIMIT stop reason."
    TOP_K = "The number of highest probability vocabulary tokens to keep for top-k-filtering. The range is any value >= 1. Valid only with decoding_method=sample."
    TOP_P = "If set to value < 1, only the most probable tokens with probabilities that add up to top_p or higher are kept for generation. The range is 0.00 to 1.00. Valid only with decoding_method=sample."
    TYPICAL_P = "Local typicality measures how similar the conditional probability of predicting a target token next is to the expected conditional probability of predicting a random token next, given the partial text already generated. If set to float < 1, the smallest set of the most locally typical tokens with probabilities that add up to typical_p or higher are kept for generation."
    REPETITION_PENALTY = "The parameter for repetition penalty. 1.0 means no penalty."
    TRUNCATE_INPUT_TOKENS = "Truncate to this many input tokens. Can be used to avoid requests failing due to input being longer than configured limits. Zero means don't truncate."
    BEAM_WIDTH = "Multiple output sequences of tokens are generated, using your decoding selection, and then the output sequence with the highest overall probability is returned. When beam search is enabled, there will be a performance penalty, and Stop sequences will not be available."  # noqa
    INCLUDE_STOP_SEQUENCE = "Indicates whether the stop sequence (which caused the generation to stop) is part of the generated text. The default value depends on the model in use."

    # Params.Token
    RETURN_TOKEN = "Return tokens with the response. Defaults to false."

    # Params.History
    LIMIT = "Specifies the maximum number of items in the collection that should be returned. Defaults to 100. Maximum is 100."
    OFFSET = "Specifies the starting position in the collection. Defaults to 0."
    STATUS = "Filters the items to be returned based on their status. Possible values are SUCCESS and ERROR."
    ORIGIN = "Filters the items to be returned based on their origin. Possible values are API and UI."

    # Params.Moderations
    MODERATIONS = "Leverages various models to detect hate speech in the provided inputs and generated outputs."
    HAP = "Mechanism for detecting hate/abuse/profanity on a sentence level."
    HAP_INPUT = "Enable/Disable HAP detection on the provided input."
    HAP_OUTPUT = "Enable/Disable HAP detection on the generated output."
    HAP_THRESHOLD = "The number from interval <0, 1> that causes the sentence to be flagged (default is 0.75)."


class TunesAPIDescriptions:
    # Params.CreateTune
    NAME = "Name of the tune."
    SEARCH = "Filters the items to be returned based on their name."
    MODEL_ID = "The ID of the model to be used for this request."
    METHOD_ID = "The ID of the tuning method to be used for this request."
    TASK_ID = "Task ID that determines format of the training data. Possible values are generic, classification, or summarization."
    TRAINING_FILE_IDS = "The IDs of uploaded files that contain training data."
    VALIDATION_FILE_IDS = "The IDs of uploaded files that contain validation data."
    PARAMETERS = "key-value pairs"
    ACCUMULATE_STEPS = "Number of training steps to use to combine gradients. This helps overcome the limitation of smaller batch sizes due to GPU memory limitations. The range is 1 to 128, defaults to 16."
    BATCH_SIZE = "The number of samples to work through before updating the internal model parameters. Optimal batch size set points are based on a combination of the number of examples you’ve uploaded as well as other parameters. If you’ve uploaded a smaller training data set, you should set your batch size lower. The range is 1 to 16, defaults to 16."
    LEARNING_RATE = "Learning rate to be used while tuning prompt vectors. The range is 0.01 to 0.5, defaults to 0.3."
    MAX_INPUT_TOKENS = "The maximum number of tokens that are accepted in the input field for each example. If any of the input rows in your training data set exceed this value, the input data will be truncated at the set maximum value. The range is 1 to 256, defaults to 256."
    MAX_OUTPUT_TOKENS = "The maximum number of tokens that are accepted in the output field for each example. If any of the output rows in your training data set exceed this value, the output data will be truncated at the set maximum value. The range is 1 to 128, defaults to 128."
    NUM_EPOCHS = "The number of times to cycle through the training data set. If you have a large training data set, a high number of epochs will take a very long time to finish tuning. The range is 1 to 50, defaults to 20."
    NUM_VIRTUAL_TOKENS = "Number of virtual tokens to be used for training. This is purely experimental. If the default value doesn’t provide good results, you may want to try selecting another value. Possible values are 20, 50, or 100, defaults to 100."
    VERBALIZER = "Verbalizer template to be used for formatting data at train and inference time. This template may use double brackets to indicate where fields from training data should be rendered. The template can contain one or both of {{input}} and {{output}}. Defaults to {{input}}."
    OFFSET = "Specifies the starting position in the collection. Defaults to 0."
    LIMIT = "Specifies the maximum number of items in the collection that should be returned. Defaults to 100. Maximum is 100."
    STATUS = "Filters the items to be returned based on their status. Possible values are: INITIALIZING, NOT_STARTED, PENDING, HALTED, RUNNING, QUEUED, COMPLETED, FAILED."
    INIT_METHOD = "Initialization method to be used. Possible values are RANDOM or TEXT. Defaults to RANDOM. Used only if the method_id is 'pt' = Prompt Tuning."
    INIT_TEXT = "Initialization text to be used. This is only applicable if init_method == TEXT. Used only if the method_id is 'pt' = Prompt Tuning."
    ID = "The ID of the tune."
    CONTENT = "The name of the asset. Available options are encoder and logs."


class FilesAPIDescriptions:
    OFFSET = "Specifies the starting position in the collection. Defaults to 0."  # noqa
    LIMIT = "Specifies the maximum number of items in the collection that should be returned. Defaults to 100. Maximum is 100."  # noqa
    SEARCH = "Filters the items to be returned based on their name."
    PURPOSE = "The intended purpose of the uploaded document. Currently only tune or template are supported."  # noqa
    TASK_ID = "Task ID that determins format of the training data. Possible values are generation, classification, or summarization. This field is required if purpose == 'tune'."  # noqa
    FILE = "The file to be uploaded."



from genai.routers.prompt_template import PromptTemplateRouter

from genai.routers.files import FilesRouter  # isort:skip
from genai.routers.tunes import TunesRouter  # isort:skip

__all__ = ["FilesRouter", "PromptTemplateRouter", "TunesRouter"]



from typing import Any

from genai.utils.watsonx_helpers import watsonx_payload


class Options:
    def __init__(
        self,
        watsonx_template=None,
        watsonx_data=None,
        watsonx_files: list = None,
        *args,
        **kwargs,
    ) -> None:
        self.d = {}
        if watsonx_template:
            self.d["template"] = watsonx_payload(watsonx_template, watsonx_data, watsonx_files)
        self.d.update(kwargs)

    def __getitem__(self, key) -> Any:
        return self.d[key]

    def keys(self):
        return self.d.keys()



from unittest import TestCase
from unittest.mock import MagicMock, patch

import pytest
from pydantic import ValidationError

from genai.schemas import GenerateParams
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestGenerateService:
    def setup_method(self):
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def params(self):
        return GenerateParams(decoding_method="sample", max_new_tokens=5, min_new_tokens=0)

    @patch("genai.services.RequestHandler.post")
    def test_generate_with_params(self, mocked_post_request: MagicMock, params):
        expected_resp = SimpleResponse.generate(model=self.model, inputs=self.inputs, params=params)
        expected = MagicMock(status_code=200, json=expected_resp)

        mocked_post_request.return_value = expected
        g = self.service.generate(model=self.model, inputs=self.inputs, params=params)

        assert g == expected
        assert g.status_code == 200
        TestCase().assertDictEqual(g.json, expected_resp)

    @patch("genai.services.RequestHandler.post")
    def test_generate_with_no_params(self, mocked_post_request):
        expected_resp = SimpleResponse.generate(model=self.model, inputs=self.inputs)
        expected = MagicMock(status_code=200, json=expected_resp)

        mocked_post_request.return_value = expected
        g = self.service.generate(model=self.model, inputs=self.inputs)

        assert g == expected
        assert g.status_code == 200
        TestCase().assertDictEqual(g.json, expected_resp)

    def test_generate_with_invalid_params(self):
        with pytest.raises(ValidationError):
            self.service.generate(
                model=self.model,
                inputs=self.inputs,
                params=GenerateParams(decoding_method="not the right method"),
            )

    def test_generate_empty_body(self):
        with pytest.raises(TypeError):
            self.service.generate()

    def test_generate_with_no_model(self):
        with pytest.raises(TypeError):
            self.service.generate(inputs=self.inputs)

    def test_generate_with_no_input(self):
        with pytest.raises(TypeError):
            self.service.generate(model=self.model)

    @patch(
        "genai.services.RequestHandler.post",
        side_effect=Exception("some general error"),
    )
    def test_generate_general_exception(self, mock):
        with pytest.raises(BaseException, match="some general error"):
            self.service.generate(model="somemodel", inputs=["inputs"])



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.prompt_pattern import PromptPattern
from genai.schemas import TokenParams
from genai.services.prompt_template_manager import PromptTemplateManager

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_url = os.getenv("GENAI_API", None)

print("\n------------- Example Mustaches Prompt [ CREATE ] -------------\n")

params = TokenParams(return_tokens=True)

creds = Credentials(api_key, api_endpoint=api_url)

_template = """
{{ instruction }}
{{#examples}}

{{input}}
{{output}}

{{/examples}}
{{input}}
"""

print("\n------------- Example Mustaches Prompt [ SAVE ] -------------\n")
test_pt = PromptPattern.from_watsonx(credentials=creds, name="test", template=_template)
print("\nSaved template information:\n", test_pt)


print("\n------------- Example Mustaches Prompt [ LIST ] -------------\n")
pts = PromptTemplateManager.load_all_templates(credentials=creds)
for r in pts.results:
    print(f"{r.name}, {r.id}")


print("\n------------- Example Mustaches Prompt [ GET ONE ] -------------\n")
template = PromptPattern.from_watsonx(credentials=creds, name="aa", template=_template)
print(template.watsonx)


print("\n------------- Example Mustaches Prompt [ UPDATE ] -------------\n")
template = PromptPattern.from_watsonx(credentials=creds, name="aa", template="{{a}}{{b}}")
print(template.watsonx)


print("\n------------- Example Mustaches Prompt [ DELETE ] -------------\n")
_id = test_pt.delete()
print(f"\n Deleted prompt with id : {_id}")



from typing import Literal, Optional, Union
from warnings import warn

from pydantic import BaseModel, Extra, Field

from genai.schemas import Descriptions as tx

# API Reference : https://workbench.res.ibm.com/docs


class LengthPenalty(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    decay_factor: Optional[float] = Field(None, description=tx.DECAY_FACTOR, gt=1.00)
    start_index: Optional[int] = Field(None, description=tx.START_INDEX)


class ReturnOptions(BaseModel):
    class Config:
        anystr_strip_whitespace = False
        extra = Extra.forbid

    input_text: Optional[bool] = Field(None, description=tx.INPUT_TEXT)
    generated_tokens: Optional[bool] = Field(None, description=tx.GENERATED_TOKEN)
    input_tokens: Optional[bool] = Field(None, description=tx.INPUT_TOKEN)
    token_logprobs: Optional[bool] = Field(None, description=tx.TOKEN_LOGPROBS)
    token_ranks: Optional[bool] = Field(None, description=tx.TOKEN_RANKS)
    top_n_tokens: Optional[int] = Field(None, description=tx.TOP_N_TOKENS)


class Return(ReturnOptions):
    def __init__(self, *args, **kwargs):
        warn(DeprecationWarning(f"{self.__class__.__name__} is deprecated, please use ReturnOptions instead."))
        super().__init__(*args, **kwargs)


# NOTE - The "return" parameter is deprecated, please use return_options now.
# Context   : The GENAI Service endpoint has an optional parameter named "return".
# Issue     : "return" is a reserved keyword, so we can't directly use it as an
#             attribute of Generate.
# Fix       : We created a "returns" attribute which gets mapped to the "return"
#             dictionary key in the sanitize method of ServiceInterface.
# Link to doc : https://workbench.res.ibm.com/docs/api-reference#generate


class HAPOptions(BaseModel):
    input: bool = Field(description=tx.HAP_INPUT, default=True)
    output: bool = Field(description=tx.HAP_OUTPUT, default=True)
    threshold: float = Field(description=tx.HAP_THRESHOLD, ge=0, le=1, multiple_of=0.01, default=0.75)


class ModerationsOptions(BaseModel):
    class Config:
        extra = Extra.allow
        allow_population_by_field_name = True

    hap: Union[bool, HAPOptions] = Field(description=tx.HAP, default=False)


class GenerateParams(BaseModel):
    class Config:
        anystr_strip_whitespace = False
        extra = Extra.allow
        allow_population_by_field_name = True

    decoding_method: Optional[Literal["greedy", "sample"]] = Field(None, description=tx.DECODING_METHOD)
    length_penalty: Optional[LengthPenalty] = Field(None, description=tx.LENGTH_PENALTY)
    max_new_tokens: Optional[int] = Field(None, description=tx.MAX_NEW_TOKEN, ge=1)
    min_new_tokens: Optional[int] = Field(None, description=tx.MIN_NEW_TOKEN, ge=0)
    random_seed: Optional[int] = Field(None, description=tx.RANDOM_SEED, ge=1)
    stop_sequences: Optional[list[str]] = Field(None, description=tx.STOP_SEQUENCES, min_length=1)
    stream: Optional[bool] = Field(None, description=tx.STREAM)
    temperature: Optional[float] = Field(None, description=tx.TEMPERATURE, ge=0.05, le=2.00)
    time_limit: Optional[int] = Field(None, description=tx.TIME_LIMIT)
    top_k: Optional[int] = Field(None, description=tx.TOP_K, ge=1)
    top_p: Optional[float] = Field(None, description=tx.TOP_P, ge=0.00, le=1.00)
    typical_p: Optional[float] = Field(None, description=tx.TYPICAL_P, gt=0.00, le=1.00)
    repetition_penalty: Optional[float] = Field(
        None, description=tx.REPETITION_PENALTY, multiple_of=0.01, ge=1.00, le=2.00
    )
    truncate_input_tokens: Optional[int] = Field(None, description=tx.TRUNCATE_INPUT_TOKENS, ge=0)
    beam_width: Optional[int] = Field(None, description=tx.BEAM_WIDTH, ge=0)
    return_options: Optional[ReturnOptions] = Field(None, description=tx.RETURN)
    returns: Optional[Return] = Field(None, description=tx.RETURN, alias="return", deprecated=True)
    moderations: Optional[ModerationsOptions] = Field(None, description=tx.MODERATIONS)
    include_stop_sequence: Optional[bool] = Field(None, description=tx.INCLUDE_STOP_SEQUENCE)



import logging
from typing import Optional

from httpx import Response
from httpx_sse import SSEError, connect_sse

from genai._version import version
from genai.exceptions import GenAiException
from genai.options import Options
from genai.services.connection_manager import ConnectionManager
from genai.utils.http_provider import HttpProvider

logger = logging.getLogger(__name__)

__all__ = ["RequestHandler"]


class RequestHandler:
    @staticmethod
    def _metadata(
        method: str,
        key: str,
        model_id: str = None,
        inputs: list = None,
        parameters: dict = None,
        options: Options = None,
        files: dict = None,
    ):
        """General function to build header and/or json_data for /post and /get requests.

        Args:
            method (str): Request type. Currently accepts GET or POST
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
            files (dict, optional): Pre-built files. Defaults to None.

        Returns:
            tuple: Headers, json_data for request, files
        """

        headers = {
            "Authorization": f"Bearer {key}",
            "x-request-origin": f"python-sdk/{version}",
            "user-agent": f"python-sdk/{version}",
        }

        # NOTE: discuss with team if we want to keep like this or try another approach
        if method == "POST" and files is not None:
            return headers, None, files

        json_data = {}

        if method == "POST" or method == "PUT":
            headers["Content-Type"] = "application/json"

            if model_id is not None:
                json_data["model_id"] = model_id

            if inputs is not None:
                json_data["inputs"] = inputs

            if parameters is not None:
                json_data["parameters"] = parameters

            if options is not None:
                for key in options.keys():
                    json_data[key] = options[key]

        if method == "PATCH":
            headers["Content-Type"] = "application/json"

        return headers, json_data, files

    @staticmethod
    async def async_post(
        endpoint: str,
        key: str,
        model_id: str = None,
        inputs: list = None,
        parameters: dict = None,
        options: Options = None,
        files: dict = None,
    ):
        """Low level API for async /post request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
            files (dict, optional): Files to be sent to the server. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, json_data, files = RequestHandler._metadata(
            method="POST",
            key=key,
            model_id=model_id,
            inputs=inputs,
            parameters=parameters,
            options=options,
            files=files,
        )
        async with HttpProvider.get_async_client() as client:
            return await client.post(endpoint, headers=headers, json=json_data, files=files)

    @staticmethod
    async def async_patch(endpoint: str, key: str, json_data: dict) -> Response:
        """Low level API for /patch request to REST API.

        Currently only used for TOU endpoint.

        Args:
            endpoint (str):
            key (str)
            json_data: (dict)

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, _, _ = RequestHandler._metadata(method="PATCH", key=key)
        async with HttpProvider.get_async_client(timeout=ConnectionManager.TIMEOUT) as client:
            return await client.patch(endpoint, headers=headers, json=json_data)

    @staticmethod
    async def async_generate(
        endpoint: str,
        key: str,
        model_id: str = None,
        inputs: list = None,
        parameters: dict = None,
        options: Options = None,
    ):
        """Low level API for async /generate request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, json_data, _ = RequestHandler._metadata(
            method="POST",
            key=key,
            model_id=model_id,
            inputs=inputs,
            parameters=parameters,
            options=options,
        )
        return await ConnectionManager.async_generate_client.post(endpoint, headers=headers, json=json_data)

    @staticmethod
    async def async_tokenize(
        endpoint: str,
        key: str,
        model_id: str = None,
        inputs: list = None,
        parameters: dict = None,
        options: Options = None,
    ):
        """Low level API for async /tokenize request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, json_data, _ = RequestHandler._metadata(
            method="POST",
            key=key,
            model_id=model_id,
            inputs=inputs,
            parameters=parameters,
            options=options,
        )
        return await ConnectionManager.async_tokenize_client.post(endpoint, headers=headers, json=json_data)

    @staticmethod
    async def async_get(endpoint: str, key: str, parameters: dict = None):
        """Low level API for async /get request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, _, _ = RequestHandler._metadata(method="GET", key=key)

        async with HttpProvider.get_async_client(timeout=ConnectionManager.TIMEOUT) as client:
            response = await client.get(url=endpoint, headers=headers, params=parameters)
        return response

    @staticmethod
    def post(
        endpoint: str,
        key: str,
        model_id: str = None,
        inputs: list = None,
        parameters: dict = None,
        streaming: bool = False,
        options: Options = None,
        files: dict = None,
    ):
        """Low level API for /post request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            model_id (str, optional): The id of the language model to be queried. Defaults to None.
            inputs (list, optional): List of inputs to be queried. Defaults to None.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
            streaming (bool, optional): Streaming response flag. Defaults to False.
            files (dict, optional): Files to be sent to the server. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
            or
            Generator of streamed response payloads from the REST API.
        """
        headers, json_data, files = RequestHandler._metadata(
            method="POST",
            key=key,
            model_id=model_id,
            inputs=inputs,
            parameters=parameters,
            options=options,
            files=files,
        )

        if streaming:
            return RequestHandler.post_stream(endpoint=endpoint, headers=headers, json_data=json_data, files=files)
        else:
            with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as s:
                response = s.post(url=endpoint, headers=headers, json=json_data, files=files)
                return response

    @staticmethod
    def patch(endpoint: str, key: str, json_data: dict) -> Response:
        """Low level API for /patch request to REST API.

        Currently only used for TOU endpoint.

        Args:
            endpoint (str):
            key (str)
            json_data: (dict)

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, _, _ = RequestHandler._metadata(method="PATCH", key=key)

        with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as s:
            response = s.patch(url=endpoint, headers=headers, json=json_data)
            return response

    @staticmethod
    def post_stream(endpoint, headers, json_data, files):
        with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as client:
            with connect_sse(
                client,
                method="POST",
                url=endpoint,
                headers=headers,
                json=json_data,
                files=files,
            ) as event_source:
                try:
                    for sse in event_source.iter_sse():
                        yield sse.data
                except SSEError as e:
                    response: Response = event_source.response
                    if "application/json" in response.headers["content-type"]:
                        response.read()
                        raise GenAiException(response)
                    raise e

    @staticmethod
    def get(endpoint: str, key: str, parameters: Optional[dict] = None) -> Response:
        """Low level API for /get request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.

        Returns:
            httpx.Response: Response from the REST API.
        """
        headers, _, _ = RequestHandler._metadata(method="GET", key=key)
        with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as s:
            response = s.get(url=endpoint, headers=headers, params=parameters)
            return response

    @staticmethod
    def put(endpoint: str, key: str, options: Optional[Options] = None) -> Response:
        """Low level API for /get request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            requests.models.Response: Response from the REST API.
        """
        headers, json_data, _ = RequestHandler._metadata(method="PUT", key=key, options=options)
        with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as s:
            response = s.put(url=endpoint, headers=headers, json=json_data)
            return response

    @staticmethod
    def delete(endpoint: str, key: str, parameters: Optional[dict] = None) -> Response:
        """Low level API for /get request to REST API.

        Args:
            endpoint (str): Remote endpoint to be queried.
            key (str): API key for authorization.
            parameters (dict, optional): Key-value pairs for model parameters. Defaults to None.

        Returns:
            requests.models.Response: Response from the REST API.
        """
        headers, _, _ = RequestHandler._metadata(method="DELETE", key=key)
        with HttpProvider.get_client(timeout=ConnectionManager.TIMEOUT) as s:
            response = s.delete(url=endpoint, headers=headers, params=parameters)
            return response



from unittest.mock import MagicMock, patch

import pytest
from pytest_httpx import HTTPXMock

from genai import Credentials, Model
from genai.exceptions import GenAiException
from genai.schemas import GenerateParams
from genai.schemas.responses import (
    GenerateResponse,
    ModelCard,
    ModelList,
    TokenizeResponse,
)
from genai.schemas.tunes_params import CreateTuneHyperParams
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestModel:
    def setup_method(self):
        self.creds = Credentials(api_key="API_KEY", api_endpoint="SERVICE_URL")
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def credentials(self):
        return Credentials("GENAI_APY_KEY")

    @pytest.fixture
    def params(self):
        return GenerateParams(
            decoding_method="sample",
            max_new_tokens=1,
            min_new_tokens=1,
            stream=False,
            temperature=0.7,
            top_k=50,
            top_p=1,
        )

    @pytest.fixture
    def prompts(self):
        return ["a", "b", "c", "d", "e"]

    @patch("genai.services.RequestHandler.post")
    def test_generate(self, mocked_post_request, credentials, params, prompts):
        """Tests that we can call the generate endpoint"""

        GENERATE_RESPONSE = SimpleResponse.generate(model="google/flan-ul2", inputs=prompts, params=params)
        expected_generated_response = GenerateResponse(**GENERATE_RESPONSE)

        response = MagicMock(status_code=200)
        response.json.return_value = GENERATE_RESPONSE
        mocked_post_request.return_value = response

        model = Model("google/flan-ul2", params=params, credentials=credentials)

        responses = model.generate_as_completed(prompts=prompts)
        responses_list = list(responses)

        assert responses_list == expected_generated_response.results
        for i, response in enumerate(responses_list):
            assert response.input_text == prompts[i]

    @patch("genai.services.RequestHandler.post")
    def test_generate_throws_exception_for_non_200(self, mock_service_generate, credentials, params, prompts):
        """Tests that the GenAiException is thrown if the status code is not 200"""

        mock_service_generate.return_value = MagicMock(status_code=500)

        model = Model("google/flan-ul2", params=params, credentials=credentials)

        with pytest.raises(GenAiException):
            model.generate(prompts=prompts)

    @patch(
        "genai.services.RequestHandler.post",
        side_effect=Exception("some general error"),
    )
    def test_generate_throws_exception_for_generic_exception(self, credentials, params, prompts, httpx_mock: HTTPXMock):
        """Tests that the GenAiException is thrown if a generic Exception is raised"""
        httpx_mock.add_response()
        model = Model("google/flan-ul2", params=params, credentials=credentials)

        with pytest.raises(GenAiException):
            model.generate(prompts=prompts)

    @patch("genai.services.RequestHandler.post")
    def test_tokenize(self, mocked_post_request, credentials, params):
        """Tests that we can call the tokenize endpoint"""

        TOKENIZE_RESPONSE = SimpleResponse.tokenize(model="google/flan-ul2", inputs=["a", "b", "c"])
        expected_token_response = TokenizeResponse(**TOKENIZE_RESPONSE)

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = TOKENIZE_RESPONSE
        mocked_post_request.return_value = mock_response

        model = Model("google/flan-ul2", params=params, credentials=credentials)

        responses = model.tokenize(["a", "b", "c"], False)

        assert responses == expected_token_response.results

    @patch("genai.services.RequestHandler.post")
    def test_create_tune(self, mock_requests, credentials):
        label = "test_label"
        model_id = self.model
        hyperparams = CreateTuneHyperParams(num_epochs=10)
        expected_response = SimpleResponse.create_tune(model_id=model_id, name=label)

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = expected_response
        mock_requests.return_value = mock_response

        base_model = Model(self.model, params=None, credentials=credentials)
        tuned_model = base_model.tune(
            name=label,
            method="mpt",
            task="generation",
            training_file_ids=["id1"],
            hyperparameters=hyperparams,
        )
        assert tuned_model.model == expected_response["results"]["id"]

    @patch("genai.services.RequestHandler.get")
    def test_status(self, mock_requests, credentials):
        label = "test_label"
        model_id = self.model
        tune_id = SimpleResponse.get_tune(model_id=model_id, name=label)["results"]["id"]

        tuned_model = Model(tune_id, params=None, credentials=credentials)

        expected_response = SimpleResponse.get_tune()
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = expected_response
        mock_requests.return_value = mock_response

        assert tuned_model.status() == expected_response["results"]["status"]

    @patch("genai.services.RequestHandler.get")
    def test_info(self, mock_requests, credentials):
        model_id = "google/flan-t5-xl"
        response = SimpleResponse.models()
        card = [m for m in response["results"] if m["id"] == model_id]
        info = ModelCard(**card[0])

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = response
        mock_requests.return_value = mock_response

        model = Model(model_id, params=None, credentials=credentials)
        assert info == model.info()

    @patch("genai.services.RequestHandler.get")
    def test_models(self, mock_requests, credentials):
        response = SimpleResponse.models()

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = response
        mock_requests.return_value = mock_response

        assert Model.models(service=self.service) == ModelList(**response).results

    @patch("genai.services.RequestHandler.get")
    def test_available(self, mock_requests, credentials):
        response = SimpleResponse.models()

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = response
        mock_requests.return_value = mock_response

        model_id = "google/flan-t5-xl"
        model = Model(model_id, params=None, credentials=credentials)
        assert model.available() is True

        model_id = "random"
        model = Model(model_id, params=None, credentials=credentials)
        assert model.available() is False

    @pytest.mark.parametrize("generate_params", [([],), (1,), ("",), (CreateTuneHyperParams())])
    def test_invalid_generate_parameters(self, generate_params, credentials):
        with pytest.raises(ValueError):
            model = Model("google/flan-ul2", params=generate_params, credentials=credentials)
            model.generate(prompts=[])



from genai.exceptions import GenAiException
from genai.options import Options
from genai.services.request_handler import RequestHandler
from genai.utils.errors import to_genai_error


class PromptTemplateRouter:
    PROMPT_TEMPLATES = "/v1/prompt_templates"

    def __init__(self, service_url: str, api_key: str) -> None:
        self.service_url = service_url.rstrip("/")
        self.key = api_key

    def prompt_output(self, inputs, template):
        try:
            endpoint = self.service_url + PromptTemplateRouter.PROMPT_TEMPLATES + "/output"
            return RequestHandler.post(
                endpoint,
                key=self.key,
                options=Options(inputs=inputs, template=template),
            )
        except Exception as e:
            raise to_genai_error(e)

    def prompt_templates(self, name, value):
        try:
            endpoint = self.service_url + PromptTemplateRouter.PROMPT_TEMPLATES
            return RequestHandler.post(endpoint, key=self.key, options=Options(name=name, value=value))
        except Exception as e:
            raise to_genai_error(e)

    def update_prompt_templates(self, id: str, name: str, value: str):
        try:
            endpoint = self.service_url + PromptTemplateRouter.PROMPT_TEMPLATES + "/" + id
            return RequestHandler.put(endpoint, key=self.key, options=Options(name=name, value=value))
        except Exception as e:
            raise to_genai_error(e)

    def get_prompt_templates(self, id: str = None):
        try:
            endpoint = self.service_url + PromptTemplateRouter.PROMPT_TEMPLATES
            endpoint = endpoint if id is None else endpoint + "/" + id

            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def delete_prompt_templates(self, id: str):
        if id is None:
            raise GenAiException("ID cannot be None when attempting to delete a prompt.")
        try:
            endpoint = self.service_url + PromptTemplateRouter.PROMPT_TEMPLATES + "/" + id
            return RequestHandler.delete(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)



import logging

import pytest

from genai.schemas.responses import (
    GenerateResponse,
    GenerateResult,
    HistoryResponse,
    HistoryResult,
    HistoryResultRequest,
    TermsOfUse,
    TokenizeResponse,
    TokenizeResult,
)
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse

logger = logging.getLogger(__name__)


@pytest.mark.unit
class TestSchemas:
    def setup_method(self):
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    def _validate_extra_field(self, input_dict, model_type):
        """

        Args:
            input_dict (_type_): _description_
            model_type (_type_): _description_

        Returns:
            bool: _description_
        """
        logger.debug("Checking that {model_type.__name__} accepts extra parameters correctly")
        # Add extra field to input dict
        input_dict["extra_field"] = "this_is_an_extra_field"
        result = model_type(**input_dict)

        assert (
            f"{model_type.__name__}" in self.caplog.text
        ), f"Warning for {model_type.__name__} should be raised for extra field"
        assert (
            result.extra_field == "this_is_an_extra_field"
        ), f"Extra field in {model_type.__name__} should be accessable and set correctly"

    def test_log_warning_thrown_on_extra_field(self, caplog):
        """This test checks that the schemas within genai.schemas.responses
        all raise a logging WARNING if an extra field is detected, whilst also
        allowing said extra fields through into the model output.
        Args:
            caplog (Generator[LogCaptureFixture, None, None]): The Logging capture during the test
        """
        self.caplog = caplog
        self.caplog.set_level(logging.DEBUG)

        # Test Generate Result
        generate_example = SimpleResponse.generate(model=self.model, inputs=self.inputs)
        self._validate_extra_field(generate_example, GenerateResponse)
        self._validate_extra_field(generate_example["results"][0], GenerateResult)

        tokenize_example = SimpleResponse.tokenize(model=self.model, inputs=self.inputs)
        self._validate_extra_field(tokenize_example, TokenizeResponse)
        self._validate_extra_field(tokenize_example["results"][0], TokenizeResult)

        history_example = SimpleResponse.history()
        self._validate_extra_field(history_example, HistoryResponse)
        self._validate_extra_field(history_example["results"][0], HistoryResult)
        self._validate_extra_field(history_example["results"][0]["request"], HistoryResultRequest)

        tou_example = SimpleResponse.terms_of_use()
        self._validate_extra_field(tou_example, TermsOfUse)



import logging

from genai._version import __version__ as v
from genai.credentials import Credentials
from genai.metadata import Metadata
from genai.model import Model
from genai.options import Options
from genai.prompt_pattern import PromptPattern

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ["Model", "Credentials", "Metadata", "PromptPattern", "Options"]

__version__ = v



import json


def json_load(file):
    """
    Loads json or jsonl data from a file

    Args:
        file (IO) : The json or jsonl file

    Returns:
        data : Python dictionary
    """
    if file.name.endswith(".jsonl"):
        data = [json.loads(line) for line in file]
    else:
        data = json.load(file)
    return data


def json_extract(obj, key, join: bool = False):
    """
    Extract nested values from a JSON tree.

    Args:
        obj : JSON tree where we are looking for the key
        key (string): The kwy we are looking for
        join (bool, optional): If True, the values are returned as a string. Defaults to False.

    Returns:
        string | list : Values for the given key. Can be returned as string or list of string
    """
    arr = []

    def extract(obj, arr, key):
        """Recursively search for values of key in JSON tree."""
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, (dict, list)):
                    extract(v, arr, key)
                elif k == key:
                    arr.append(v)
        elif isinstance(obj, list):
            for item in obj:
                extract(item, arr, key)
        return arr

    values = extract(obj, arr, key)
    return "".join(values) if join else values


def json_get_all_keys(obj, join: bool = False):
    arr = []

    def extract_keys(obj, arr):
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, (dict, list)):
                    extract_keys(v, arr)
                else:
                    arr.append(k)
        elif isinstance(obj, list):
            for item in obj:
                extract_keys(item, arr)
        return arr

    keys = extract_keys(obj, arr)
    return "".join(keys) if join else keys



from genai.model import TokenParams
from genai.schemas.generate_params import GenerateParams
from genai.schemas.responses import GenAiResponseModel


class GenerateRequestBody(GenAiResponseModel):
    model_id: str
    inputs: list[str]
    parameters: GenerateParams


class TokenizeRequestBody(GenAiResponseModel):
    model_id: str
    inputs: list[str]
    parameters: TokenParams



import asyncio
from typing import List, Optional

import httpx
from aiolimiter import AsyncLimiter
from httpx import HTTPStatusError, Request, RequestError, Response

__all__ = ["AsyncRateLimiter", "AsyncRateLimitTransport", "AsyncRetryTransport"]


class AsyncRateLimiter(AsyncLimiter):
    def update_limit(self, *, max_rate: Optional[float] = None, time_period: Optional[float] = None):
        self.max_rate = max_rate or self.max_rate
        self.time_period = time_period or self.time_period
        self._rate_per_sec = self.max_rate / self.time_period


class AsyncRetryTransport(httpx.AsyncHTTPTransport):
    def __init__(self, *args, retry_status_codes: Optional[List[int]] = None, backoff_factor: float = 0.2, **kwargs):
        self.retry_status_codes = retry_status_codes or [
            httpx.codes.TOO_MANY_REQUESTS,
            httpx.codes.BAD_GATEWAY,
            httpx.codes.SERVICE_UNAVAILABLE,
        ]
        self.backoff_factor = backoff_factor
        self.retries = kwargs.get("retries", 0)
        super().__init__(*args, **kwargs)

    def _get_retry_delays(self):
        yield 0
        for i in range(self.retries):
            yield self.backoff_factor * (2**i)

    async def handle_async_request(
        self,
        request: Request,
    ) -> Response:
        latest_err: Optional[Exception] = None

        for delay in self._get_retry_delays():
            if delay > 0:
                await asyncio.sleep(delay)

            try:
                response = await super().handle_async_request(request)
                response.request = request
                response.raise_for_status()
                return response
            except HTTPStatusError as ex:
                latest_err = ex
                if ex.response.status_code in self.retry_status_codes:
                    continue
                raise ex

        raise RequestError(f"Failed to handle request to {request.url}", request=request) from latest_err


class AsyncRateLimitTransport(AsyncRetryTransport):
    def __init__(self, *args, default_max_rate: float, default_time_period: float, **kwargs):
        super().__init__(*args, **kwargs)
        self.limiter = AsyncRateLimiter(max_rate=default_max_rate, time_period=default_time_period)

    def update_rate_limit(self, response: Response):
        max_capacity = response.headers.get("x-ratelimit-limit")
        reset_period_in_seconds = response.headers.get("x-ratelimit-reset")

        if max_capacity and reset_period_in_seconds:
            self.limiter.update_limit(
                max_rate=max(1, int(max_capacity)),
                time_period=max(1, int(reset_period_in_seconds)),
            )

    async def handle_async_request(
        self,
        request: Request,
    ) -> Response:
        async with self.limiter:
            response = await super().handle_async_request(request)
            self.update_rate_limit(response)
            return response



import json
import os
import pathlib
from unittest.mock import patch

import pytest

from genai.prompt_pattern import PromptPattern


@pytest.mark.unit
class TestPromptPattern:
    def setup_method(self):
        self.path = pathlib.Path(__file__).parent.resolve()
        self.asset_path = str(self.path) + os.sep + "assets" + os.sep

    def asset(self, name):
        with open(self.asset_path + name, "r", encoding="utf-8") as file:
            return file.read()

    def test_validate_yaml(self):
        pt = PromptPattern.from_file(self.asset_path + "capital-qa.yaml")
        assert pt.validate() is True

    def test_validate_local_yaml_prompt(self):
        pt = PromptPattern.from_file(self.asset_path + "invalid-template.yaml")
        with pytest.raises(Exception) as context:
            pt.validate()
        assert "Invalid Prompt YAML. It must include 'apiVersion'." in str(context)

    def test_invalide_local_yaml_content(self):
        with pytest.raises(Exception) as context:
            PromptPattern.from_file(self.asset_path + "invalid-content-template.yaml")
        assert "Invalid Prompt YAML. It must include 'content' fields." in str(context)

    @patch("genai.PromptPattern._fetch_prompt")
    def test_validate_yaml_from_url(self, mock_yaml):
        invalid_yaml = {"content": "this is a test prompt"}
        mock_yaml.return_value = invalid_yaml

        pt = PromptPattern.from_url(url="https://yaml", token="token")
        pt.yaml = invalid_yaml

        with pytest.raises(Exception) as context:
            pt.validate()
        assert "Invalid Prompt YAML. It must include 'apiVersion'." in str(context)

    @patch("genai.PromptPattern._fetch_prompt")
    def test_validate_yaml_content_from_url(self, mock_yaml):
        invalid_yaml = {"apiVersion": "v0", "prompts": "this is a test prompt"}
        mock_yaml.return_value = invalid_yaml

        pt = PromptPattern.from_url(url="https://yaml", token="token")
        pt.yaml = invalid_yaml

        with pytest.raises(Exception) as context:
            pt.validate()
        assert "Invalid Prompt YAML. It must include 'content' fields." in str(context)

    def test_sub(self):
        pt = PromptPattern.from_file(self.asset_path + "story.yaml")
        pt.sub("company", "Star Trek Fleet Command").sub("person", "Mr.Sock")

        expected = self.asset(name="story_sub.txt")
        assert str(pt) == expected

    def test_sub_string_template(self):
        pt = PromptPattern.from_file(self.asset_path + "capital-qa.yaml")
        pt.sub("country", "United Kingdom")

        expected = self.asset(name="capital-qa-sub.txt")
        assert str(pt) == expected

    def test_cache_reset(self):
        pt = PromptPattern.from_file(self.asset_path + "capital-qa.yaml")
        pt.sub("country", "United Kingdom")

        expected = self.asset(name="capital-qa-sub.txt")
        assert str(pt) == expected
        assert str(pt) == expected

        pt.reset()

        assert str(pt) == self.asset(name="capital-qa-content.txt")

    def test_sub_with_string(self):
        prompt_content = (
            "Once upon a time at {{company}} there was a "
            "{{person}}. This {{person}} had ambitious but "
            "logical ideas."
        )

        expected = (
            "Once upon a time at Star Trek Fleet Command there was a "
            "Mr.Sock. This Mr.Sock had ambitious but "
            "logical ideas."
        )
        pt = PromptPattern.from_str(prompt_content)

        pt.sub("company", "Star Trek Fleet Command").sub("person", "Mr.Sock")

        assert str(pt) == expected

    def test_find_vars(self):
        pt = PromptPattern.from_file(self.asset_path + "capital-qa.yaml")
        vars = pt.find_vars()

        expected = set(["country"])
        assert vars == expected

        pt = PromptPattern.from_file(self.asset_path + "story.yaml")
        vars = pt.find_vars()

        expected = set(["company", "person"])
        assert vars == expected

        pt = PromptPattern.from_str("My {{stuff }} is {{name}} and {{ name}} means {{ name }}")
        assert pt.find_vars() == {"name", "stuff"}

    def test_private_constructor(self):
        with pytest.raises(TypeError) as context:
            PromptPattern("some_string")
        assert "PromptPattern() takes no arguments" in str(context)

    @patch("genai.PromptPattern._fetch_prompt")
    def test_from_url_and_refetch(self, mock):
        raw_prompt = "A template from far far away..."
        mock.return_value = raw_prompt

        pt = PromptPattern.from_url(url="https://url.yaml", token="some_token")

        assert pt.raw_content == raw_prompt
        assert pt.token == "some_token"

        pt.raw_content = "Another template from far away."
        pt.refetch()
        assert pt.raw_content == raw_prompt

    def test_reset(self):
        raw_prompt = "A template from far far away..."
        pt = PromptPattern.from_str(raw_prompt)
        pt.dump = None
        pt.reset()

        assert pt.dump == raw_prompt

    def test_sub_from_csv_random(self):
        pt = PromptPattern.from_file(self.asset_path + "csv_file.yaml")

        path = self.asset_path + "csv_file.csv"
        res = pt.sub_from_csv(
            csv_path=path,
            strategy="random",
            col_to_var={
                "col1": ["v1", "v2"],
                "col2": ["v4"],
                "col3": ["v3"],
            },
        )

        assert isinstance(res, PromptPattern) is True
        assert str(pt).replace("\n", "") == "aa,aa,cc,bb,"

    def test_sub_from_rand_row(self):
        # Test we can change only one var:col
        pt = PromptPattern.from_file(self.asset_path + "csv_file.yaml")

        path = self.asset_path + "csv_file.csv"
        pt.sub_from_csv(csv_path=path, col_to_var={"col1": ["v2"]})

        assert str(pt).replace("\n", "") == "{{v1}},aa,{{v3}},{{v4}},"

    def test_sub_from_json(self):
        with open(os.path.join(self.asset_path, "capital-qa-sub.json"), "r") as fin:
            arr = json.load(fin)
            countries = [a["pais"] for a in arr]
            capitals = [a["capitulo"] for a in arr]
        # test for sequential
        pt = PromptPattern.from_file(self.asset_path + "capital-qa-multivar.yaml")
        pt = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sequential",
            start_index=0,
            n=1,
        )
        pstr = str(pt)
        assert sum(c in pstr for c in countries) == 3
        assert sum(c in pstr for c in capitals) == 2
        pt.reset()
        pstr = str(pt)
        assert sum(c in pstr for c in countries) == 0
        assert sum(c in pstr for c in capitals) == 0
        # test for sample
        pt = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sample",
            start_index=0,
            n=1,
        )
        pstr = str(pt)
        assert sum(c in pstr for c in countries) == 3
        assert sum(c in pstr for c in capitals) == 2
        pt.reset()
        # test for random and multiple prompts
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="random",
            start_index=0,
            n=2,
        )
        assert len(ptarr) == 2
        assert isinstance(ptarr[0], PromptPattern) is True
        # test fault tolerance if n too large than what file allows
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sample",
            start_index=0,
            n=100,
        )
        assert len(ptarr) < 100
        assert isinstance(ptarr[0], PromptPattern) is True
        # test sequential does not return all n prompts if file is small
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sequential",
            start_index=0,
            n=2,
        )
        assert len(ptarr) < 2
        assert isinstance(ptarr[0], PromptPattern) is True
        # test random can return any number of prompts
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="random",
            start_index=0,
            n=100,
        )
        assert len(ptarr) == 100
        assert isinstance(ptarr[0], PromptPattern) is True
        # test with n = -1 it returns as many prompts as it can
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sequential",
            start_index=0,
            n=-1,
        )
        assert len(ptarr) == 1
        assert isinstance(ptarr[0], PromptPattern) is True
        # test for start index
        ptarr = pt.sub_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
            strategy="sequential",
            start_index=1,
            n=-1,
        )
        assert len(ptarr) == 1
        assert "Estados Unidos" not in str(ptarr[0])

    def test_sub_all_from_json(self):
        with open(os.path.join(self.asset_path, "capital-qa-sub.json"), "r") as fin:
            arr = json.load(fin)
            countries = [a["pais"] for a in arr]
            capitals = [a["capitulo"] for a in arr]
        # test for sequential
        pt = PromptPattern.from_file(self.asset_path + "capital-qa-multivar.yaml")
        ptarr = pt.sub_all_from_json(
            json_path=os.path.join(self.asset_path, "capital-qa-sub.json"),
            key_to_var={"pais": ["country1", "country2", "country3"], "capitulo": ["capital1", "capital2"]},
        )
        assert len(ptarr) == 1
        assert isinstance(ptarr[0], PromptPattern) is True
        pstr = str(ptarr[0])
        assert sum(c in pstr for c in countries) == 3
        assert sum(c in pstr for c in capitals) == 2

    def test_tabular_infer_mode_helper(self):
        first_row = ["var1", "var2", "var3"]

        pt = PromptPattern.from_str("###".join(["{{" + v + "}}" for v in first_row]))
        col_to_var, columns = pt._tabular_infer_mode_helper(first_row)
        assert col_to_var == {v: [v] for v in first_row}
        assert columns == first_row

        pt = PromptPattern.from_str("First var {{0}}, then {{1}}, then {{ 2 }}")
        col_to_var, columns = pt._tabular_infer_mode_helper(first_row)
        assert col_to_var == {v: [v] for v in ["0", "1", "2"]}
        assert columns == ["0", "1", "2"]

        pt = PromptPattern.from_str("Var {{1}}, then {{2}}")
        col_to_var, columns = pt._tabular_infer_mode_helper(first_row)
        assert col_to_var == {v: [v] for v in ["1", "2"]}
        assert columns == ["0", "1", "2"]

    def test_csv_infer_mode(self):
        pt = PromptPattern.from_str("Country {{country}} capital {{capital}}")
        plist = pt.sub_all_from_csv(csv_path=os.path.join(self.asset_path, "capital-qa-sub.csv"), col_to_var="infer")
        assert len(plist) == 3
        assert str(plist[0]) == "Country USA capital Washington"
        assert str(plist[2]) == "Country China capital Beijing"

        pt = PromptPattern.from_str("Capital {{capital}} Nonexistent {{var}}")
        plist = pt.sub_all_from_csv(csv_path=os.path.join(self.asset_path, "capital-qa-sub.csv"), col_to_var="infer")
        assert len(plist) == 3
        assert str(plist[0]) == "Capital Washington Nonexistent {{var}}"
        assert str(plist[2]) == "Capital Beijing Nonexistent {{var}}"

        pt = PromptPattern.from_str("Country {{0}} capital {{1}}")
        plist = pt.sub_all_from_csv(
            csv_path=os.path.join(self.asset_path, "capital-qa-sub-noheader.csv"), col_to_var="infer", headers=False
        )
        assert len(plist) == 3
        assert str(plist[0]) == "Country USA capital Washington"
        assert str(plist[2]) == "Country China capital Beijing"

        pt = PromptPattern.from_str("Capital {{1}} Nonexistent {{15}}")
        plist = pt.sub_all_from_csv(
            csv_path=os.path.join(self.asset_path, "capital-qa-sub-noheader.csv"), col_to_var="infer", headers=False
        )
        assert len(plist) == 3
        assert str(plist[0]) == "Capital Washington Nonexistent {{15}}"
        assert str(plist[2]) == "Capital Beijing Nonexistent {{15}}"

    def test_sub_from_csv_infer_mode(self):
        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path + "csv_file.csv"

        res = pt.sub_all_from_csv(csv_path=path, col_to_var="infer")
        assert isinstance(res, list)
        assert len(res) == 2
        assert str(res[0]) == "aa,bb,cc"
        assert str(res[1]) == "aa,bb,cc"

    def test_sub_from_csv_start_index_random_strategy(self):
        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path + "csv_file.csv"

        with pytest.raises(Exception) as e:
            pt.sub_all_from_csv(csv_path=path, col_to_var="infer", start_index=0, strategy="random")
            assert "start_index" in str(e)

    def test_sub_from_csv_start_index_sample_strategy(self):
        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path + "csv_file.csv"

        with pytest.raises(Exception) as e:
            pt.sub_all_from_csv(csv_path=path, col_to_var="infer", start_index=0, strategy="sample")
            assert "start_index" in str(e)



from unittest.mock import MagicMock, patch

import pytest

from genai import Credentials
from genai.extensions.huggingface import IBMGenAIAgent
from genai.schemas import GenerateParams
from tests.assets.response_helper import SimpleResponse


@pytest.mark.extension
class TestHuggingfaceAgent:
    @pytest.fixture
    def credentials(self):
        return Credentials("GENAI_APY_KEY")

    @pytest.fixture
    def params(self):
        return GenerateParams()

    @pytest.fixture
    def mock_model_response(self):
        return [
            """tool: `summarizer` to create a summary of the input text.
        Answer:
        ```py
        summarized_text = summarizer(text)
        print(f"Summary: {summarized_text}")
        ```

        Task:"""
        ]

    @patch("httpx.Client.post")
    def test_agent(self, mocked_post_request, credentials, params, mock_model_response):
        GENERATE_RESPONSE = SimpleResponse.generate(
            model="google/flan-ul2",
            inputs=mock_model_response,
            params=params,
        )

        response = MagicMock(status_code=200)
        response.json.return_value = GENERATE_RESPONSE
        mocked_post_request.return_value = response

        agent = IBMGenAIAgent(credentials=credentials, model="google/flan-ul2", params=params)
        agent.run("Summarize the text", text="Testing the summarization")



from typing import Optional

from pydantic import BaseModel, Extra, Field

from genai.schemas.descriptions import FilesAPIDescriptions as tx


class FileListParams(BaseModel):
    """Class to hold the parameters for file listing."""

    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    limit: Optional[int] = Field(None, description=tx.LIMIT, le=100)
    offset: Optional[int] = Field(None, description=tx.OFFSET)
    search: Optional[str] = Field(None, description=tx.SEARCH)


class MultipartFormData(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    purpose: str = Field(..., description=tx.PURPOSE)
    file: str = Field(..., description=tx.FILE)



from unittest import TestCase
from unittest.mock import MagicMock, patch

import pytest
from pydantic import ValidationError

from genai.schemas import TokenParams
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse

# API Reference : https://workbench.res.ibm.com/docs/api-reference#generate


@pytest.mark.unit
class TestTokenize:
    def setup_method(self):
        # mock object for the API call
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def request_body(self):
        return {
            "model": self.model,
            "inputs": self.inputs,
            "params": self.params,
        }

    @pytest.fixture
    def params(self):
        return TokenParams(return_tokens=True)

    @pytest.fixture
    def no_token_params(self):
        return TokenParams(return_tokens=False)

    @patch("genai.services.RequestHandler.post")
    def test_tokenize(self, mock_requests, params):
        expected_resp = SimpleResponse.tokenize(model=self.model, inputs=self.inputs, params=params)
        mock_response = MagicMock(status_code=200, json=expected_resp)

        mock_requests.return_value = mock_response
        t = self.service.tokenize(model=self.model, inputs=self.inputs, params=params)

        assert t == mock_response
        assert t.status_code == 200

    @patch("genai.services.RequestHandler.post")
    def test_no_tokens(self, mock_requests, no_token_params):
        expected_resp = SimpleResponse.tokenize(model=self.model, inputs=self.inputs, params=no_token_params)
        mock_response = MagicMock(status_code=200, json=expected_resp)

        mock_requests.return_value = mock_response
        t = self.service.tokenize(model=self.model, inputs=self.inputs, params=no_token_params)

        assert t == mock_response
        TestCase().assertDictEqual(t.json, expected_resp)

    def test_required_fields(self, request_body):
        assert "model" in request_body
        assert "inputs" in request_body

    def test_model(self, request_body):
        assert len(self.model) > 0
        assert all(isinstance(model, str) for model in self.model)

    def test_inputs(self, request_body):
        assert len(self.inputs) > 0
        assert all(isinstance(input, str) for input in self.inputs)

    def test_invalid_fields(self):
        with pytest.raises(ValidationError):
            TokenParams(return_tokens="dummy")

    @patch(
        "genai.services.RequestHandler.post",
        side_effect=Exception("some general error"),
    )
    def test_tokenize_general_exception(self, mock):
        with pytest.raises(BaseException, match="some general error"):
            self.service.tokenize(model="somemodel", inputs=["inputs"])



import pathlib

import pytest


@pytest.mark.extension
class TestPandasExtension:
    def setup_method(self):
        self.path = pathlib.Path(__file__).parent.resolve()
        self.asset_path = pathlib.Path(__file__, "..", "..", "assets").resolve()

    def test_sub_from_dataframe_random(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_file(str(self.asset_path / "csv_file.yaml"))

        path = self.asset_path / "csv_file.csv"
        df = pd.read_csv(path)

        pt.pandas.sub_from_dataframe(
            dataframe=df,
            strategy="random",
            col_to_var={
                "col1": ["v1", "v2"],
                "col2": ["v4"],
                "col3": ["v3"],
            },
        )
        assert str(pt).replace("\n", "") == "aa,aa,cc,bb,"

    def test_sub_from_dataframe_start_index_random_strategy(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path / "csv_file.csv"
        df = pd.read_csv(path)

        with pytest.raises(Exception) as e:
            pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer", start_index=0, strategy="random")
            assert "start_index" in str(e)

    def test_sub_from_dataframe_start_index_sample_strategy(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path / "csv_file.csv"
        df = pd.read_csv(path)

        with pytest.raises(Exception) as e:
            pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer", start_index=0, strategy="sample")
            assert "start_index" in str(e)

    def test_sub_from_dataframe_random_row(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_file(str(self.asset_path / "csv_file.yaml"))

        path = self.asset_path / "csv_file.csv"
        df = pd.read_csv(path)

        pt.pandas.sub_from_dataframe(dataframe=df, col_to_var={"col1": ["v2"]})
        assert str(pt).replace("\n", "") == "{{v1}},aa,{{v3}},{{v4}},"

    def test_sub_from_dataframe_infer_mode(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("{{col1}},{{col2}},{{col3}}")
        path = self.asset_path / "csv_file.csv"
        df = pd.read_csv(path)

        res = pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer")
        assert isinstance(res, list)
        assert len(res) == 2
        assert str(res[0]) == "aa,bb,cc"
        assert str(res[1]) == "aa,bb,cc"

    def test_sub_all_from_dataframe_infer_mode(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("Country {{country}} capital {{capital}}")
        path = self.asset_path / "capital-qa-sub.csv"
        df = pd.read_csv(path)

        plist = pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer")
        assert len(plist) == 3
        assert str(plist[0]) == "Country USA capital Washington"
        assert str(plist[2]) == "Country China capital Beijing"

    def test_sub_all_from_dataframe_nonexistent_var_infer_mode(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("Capital {{capital}} Nonexistent {{var}}")
        path = self.asset_path / "capital-qa-sub.csv"
        df = pd.read_csv(path)

        plist = pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer")
        assert len(plist) == 3
        assert str(plist[0]) == "Capital Washington Nonexistent {{var}}"
        assert str(plist[2]) == "Capital Beijing Nonexistent {{var}}"

    def test_sub_all_from_dataframe_with_no_header(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("Country {{0}} capital {{1}}")
        path = self.asset_path / "capital-qa-sub-noheader.csv"

        df_no_header = pd.read_csv(path, header=None)
        plist = pt.pandas.sub_all_from_dataframe(dataframe=df_no_header, col_to_var="infer")
        assert len(plist) == 3
        assert str(plist[0]) == "Country USA capital Washington"
        assert str(plist[2]) == "Country China capital Beijing"

        df = pd.read_csv(path)
        plist = pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer", headers=False)
        assert len(plist) == 3
        assert str(plist[0]) == "Country USA capital Washington"
        assert str(plist[2]) == "Country China capital Beijing"

    def test_infer_mode_col_to_var_empty(self):
        import pandas as pd

        import genai.extensions.pandas  # noqa
        from genai.prompt_pattern import PromptPattern

        pt = PromptPattern.from_str("{{animal}},{{specices}},{{island}},{{something}},{{year}}")
        path = self.asset_path / "penguins.csv"
        with pytest.raises(Exception) as e:
            df = pd.read_csv(path, index_col=0, header=None)
            pt.pandas.sub_all_from_dataframe(dataframe=df, col_to_var="infer", headers=False)
            assert "col_to_var" in str(e)



import os
from typing import Any, Optional
from uuid import UUID

from dotenv import load_dotenv
from langchain.callbacks.base import BaseCallbackHandler

from genai.credentials import Credentials
from genai.extensions.langchain import LangChainInterface
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)


class Callback(BaseCallbackHandler):
    def on_llm_new_token(
        self,
        token: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        print(f"Token received: {token}")


llm = LangChainInterface(
    model="google/flan-ul2",
    credentials=Credentials(api_key, api_endpoint),
    params=GenerateParams(
        decoding_method="sample",
        max_new_tokens=10,
        min_new_tokens=1,
        stream=True,
        temperature=0.5,
        top_k=50,
        top_p=1,
    ),
)

result = llm.generate(
    prompts=["Tell me about IBM."],
    callbacks=[Callback()],
)
print(result)



import csv
import os
import pathlib
import random

import requests

# Either import genai.extensions.langchain or import specific
# class from langchain extension for PromptInterface to register
import genai.extensions.langchain
from genai.prompt_pattern import PromptPattern

#
# Adult dataset in the following example was obtained from UCI Machine
# Learning repository. The full Bibtex citation is as follows:
#
# @misc{Dua:2019 ,
# author = "Dua, Dheeru and Graff, Casey",
# year = "2017",
# title = "{UCI} Machine Learning Repository",
# url = "http://archive.ics.uci.edu/ml",
# institution = "University of California, Irvine, School of Information and Computer Sciences" }


def get_dataset(datafile, headers=None):
    url = "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data"

    if os.path.exists(datafile):
        print("Data already downloaded. Skipping the download.")
        return
    with requests.get(url) as response:
        print("Saving data to {}".format(datafile))
        # Add header
        if headers is not None:
            with open(datafile, "w", newline="") as f:
                csv.writer(f).writerow(headers)
        # Add data
        with open(datafile, "ab") as f:
            f.write(response.content)


headers = [
    "age",
    "workclass",
    "demographic",
    "education_str",
    "education",
    "marital_status",
    "occupation",
    "relationship_status",
    "race",
    "sex",
    "capital_gain",
    "capital_loss",
    "work_hours_per_week",
    "native_country",
    "salary",
]
datafile = os.path.join(pathlib.Path(__file__).parent.absolute(), "adult.header.data.csv")
get_dataset(datafile, headers=headers)

# flake8: noqa
pattern = """
    The age of the individual is {{age}}.
    The workclass of the individual is {{workclass}}.
    The estimate of the number of individuals in the population with the same demographics as the individual was {{demographic}}.
    The numeric form of the highest education level they achieved was {{education}}.
    The string form of the highest education level they achieved was {{education_str}}.
    The marital status of the individual was {{marital_status}}.
    The occupation of the individual was {{occupation}}.
    The relationship status of the individual is {{relationship_status}}.
    The race of the individual was {{race}}.
    The sex of the individual is {{sex}}.
    The capital gain they had in the previous year was {{capital_gain}}.
    The capital loss they had in the previous year was {{capital_loss}}.
    The number of hours worked per week was {{work_hours_per_week}}.
    The native country of the individual was {{native_country}}.
    The salary of this individual was {{salary}}.
"""


prompt_pattern = PromptPattern.from_str(pattern)

print("\nGiven pattern:\n", prompt_pattern)

prompts = prompt_pattern.sub_all_from_csv(csv_path=datafile, col_to_var="infer")

print("-----------------------")
print("Generated prompts: \n total number {}".format(len(prompts)))
print("Sample prompt: {}".format(prompts[random.randint(0, len(prompts) - 1)]))
print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

# Langchain raw string templates are created as
# PromptTemplate(input_vars=[variable list], template=raw_string).
# In the following example, this template gets created automatically
# from genai PromptPattern using langchain extension.
# Langchain PromptTemplates are substituted with variable values as
# template.format(variable1=val1, variable2=val2, ....)
# In the following example, when we read csv using DictReader, each
# line gets read as a dictionary {column1: val, column2: val, ...}.
# We use ** operator in python to feed this dictionary to format
# the langchain template as template.format(**line)

print("Translating prompt pattern to langchain template")
template = prompt_pattern.langchain.as_template()
print("Generated langchain template = ", template)
prompts = []
with open(datafile, "r") as fin:
    reader = csv.DictReader(fin)
    for line in reader:
        if len(line) < len(headers):
            continue
        prompts.append(template.format(**line))  # **operator converts dict into key-val args

print("-----------------------")
print("Generated prompts: \n total number {}".format(len(prompts)))
print("Sample prompt: {}".format(prompts[random.randint(0, len(prompts) - 1)]))
print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

# In the following snippet, we create a genai PromptPattern from
# a langchain PromptTemplate using langchain extension of genai.
print("Translating langchain template back to prompt pattern")
prompt_pattern = PromptPattern.langchain.from_template(template)
prompts = prompt_pattern.sub_all_from_csv(csv_path=datafile, col_to_var="infer")
print("-----------------------")
print("Generated prompts: \n total number {}".format(len(prompts)))
print("Sample prompt: {}".format(prompts[random.randint(0, len(prompts) - 1)]))
print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")



import json
from unittest.mock import MagicMock, patch

import pytest
from langchain.callbacks.base import BaseCallbackHandler

from genai import Credentials
from genai.schemas import GenerateParams
from genai.schemas.responses import GenerateResponse
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse


@pytest.mark.extension
class TestLangChain:
    def setup_method(self):
        self.service = ServiceInterface(service_url="SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def credentials(self):
        return Credentials("GENAI_APY_KEY")

    @pytest.fixture
    def params(self):
        return GenerateParams()

    @pytest.fixture
    def prompts(self):
        return ["Hi! How's the weather, eh?"]

    @pytest.fixture
    def multi_prompts(self):
        return ["What is IBM?", "What is AI?"]

    @patch("httpx.Client.post")
    def test_langchain_interface(
        self,
        mocked_post_request,
        credentials,
        params,
        prompts,
    ):
        from genai.extensions.langchain import LangChainInterface

        GENERATE_RESPONSE = SimpleResponse.generate(model="google/flan-ul2", inputs=prompts, params=params)
        expected_generated_response = GenerateResponse(**GENERATE_RESPONSE)

        response = MagicMock(status_code=200)
        response.json.return_value = GENERATE_RESPONSE
        mocked_post_request.return_value = response

        model = LangChainInterface(model="google/flan-ul2", params=params, credentials=credentials)
        results = model(prompts[0])
        assert results == expected_generated_response.results[0].generated_text

    @pytest.mark.asyncio
    @patch("httpx.Client.post")
    async def test_async_langchain_interface(self, mocked_post_request, credentials, params, multi_prompts):
        from genai.extensions.langchain import LangChainInterface

        server_response = SimpleResponse.generate(model="google/flan-ul2", inputs=multi_prompts, params=params)
        expected_response = GenerateResponse(**server_response)
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = server_response
        mocked_post_request.return_value = mock_response

        model = LangChainInterface(model="google/flan-ul2", params=params, credentials=credentials)
        observed = await model.agenerate(prompts=multi_prompts)
        assert len(observed.generations) == 2
        assert len(observed.generations[0]) == 1
        assert len(observed.generations[1]) == 1

        assert observed.llm_output["token_usage"]["input_token_count"] == sum(
            c.input_token_count for c in expected_response.results
        )
        assert observed.llm_output["token_usage"]["generated_token_count"] == sum(
            c.generated_token_count for c in expected_response.results
        )

        for idx, generation_list in enumerate(observed.generations):
            assert len(generation_list) == 1
            generation = generation_list[0]
            assert generation.text == expected_response.results[idx].generated_text

            expected_result = expected_response.results[idx].dict()
            for key in {"generated_token_count", "input_text", "stop_reason"}:
                assert generation.generation_info[key] == expected_result[key]

    @patch("genai.services.RequestHandler.post_stream")
    def test_langchain_stream(self, mock_post_stream, credentials, params, prompts):
        from genai.extensions.langchain import LangChainInterface

        GENERATE_STREAM_RESPONSES = SimpleResponse.generate_stream(
            model="google/flan-ul2", inputs=prompts, params=params
        )

        response = MagicMock(status_code=200)
        response.__iter__.return_value = (json.dumps(response) for response in GENERATE_STREAM_RESPONSES)
        mock_post_stream.return_value = response

        callback = BaseCallbackHandler()
        callback.on_llm_new_token = MagicMock()

        model = LangChainInterface(
            model="google/flan-ul2",
            params=params,
            credentials=credentials,
            callbacks=[callback],
        )
        expected_generated_responses = [GenerateResponse(**result) for result in GENERATE_STREAM_RESPONSES]

        # Verify results
        for idx, result in enumerate(model.stream(input=prompts[0])):
            expected_result = expected_generated_responses[idx]
            assert result == expected_result.results[0].generated_text

        # Verify that callbacks were called
        assert callback.on_llm_new_token.call_count == len(expected_generated_responses)
        for idx, result in enumerate(expected_generated_responses):
            retrieved_kwargs = callback.on_llm_new_token.call_args_list[idx].kwargs
            token = retrieved_kwargs["token"]
            assert token == result.results[0].generated_text
            response = retrieved_kwargs["response"]
            assert response == result.results[0]

    def test_prompt_translator(self):
        from langchain.prompts import PromptTemplate

        import genai.extensions.langchain  # noqa
        from genai.prompt_pattern import PromptPattern

        s = "My name is {name}. I work for {company}. I live in {city}."
        pt = PromptTemplate(input_variables=["name", "company", "city"], template=s)
        pattern = PromptPattern.langchain.from_template(pt)
        assert pattern.find_vars() == {"name", "company", "city"}
        ptemp = pattern.langchain.as_template()
        assert set(ptemp.input_variables) == {"name", "company", "city"}
        assert ptemp.template == s



"""Wrapper around IBM GENAI APIs for use in Langchain"""
import asyncio
import logging
from functools import partial
from typing import Any, Iterator, List, Mapping, Optional

from langchain.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain.schema import LLMResult
from langchain.schema.output import Generation, GenerationChunk
from pydantic import BaseModel, Extra

from genai.exceptions import GenAiException
from genai.utils.general import to_model_instance

try:
    from langchain.llms.base import LLM
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")

from genai import Credentials, Model
from genai.schemas import GenerateParams

logger = logging.getLogger(__name__)

__all__ = ["LangChainInterface"]


class LangChainInterface(LLM, BaseModel):
    """
    Wrapper around IBM GENAI models.
    To use, you should have the ``genai`` python package installed
    and initialize the credentials attribute of this class with
    an instance of ``genai.Credentials``. Model specific parameters
    can be passed through to the constructor using the ``params``
    parameter, which is an instance of GenerateParams.
    Example:
        .. code-block:: python
            llm = LangChainInterface(model="google/flan-ul2", credentials=creds)
    """

    credentials: Credentials = None
    model: Optional[str] = None
    params: Optional[GenerateParams] = None

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        _params = to_model_instance(self.params, GenerateParams)
        return {
            **{"model": self.model},
            **{"params": _params},
        }

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "IBM GENAI"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Call the IBM GENAI's inference endpoint.
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
            run_manager: Optional callback manager.
        Returns:
            The string generated by the model.
        Example:
            .. code-block:: python
                llm = LangChainInterface(model_id="google/flan-ul2", credentials=creds)
                response = llm("What is a molecule")
        """
        result = self._generate(prompts=[prompt], stop=stop, run_manager=run_manager, **kwargs)
        return result.generations[0][0].text

    def _update_llm_result(self, current: LLMResult, generation_info: Optional[dict]):
        if generation_info is None:
            return

        token_usage = current.llm_output["token_usage"]
        for key in {"generated_token_count", "input_token_count"}:
            if key in generation_info:
                new_tokens = generation_info.get(key, 0) or 0
                token_usage[key] = token_usage.get(key, 0) + new_tokens

    def _create_generation_info(self, response: dict):
        keys_to_pick = {"stop_reason"}
        return {k: v for k, v in response.items() if k in keys_to_pick and v is not None}

    def _create_full_generation_info(self, response: dict):
        keys_to_omit = {"generated_text"}
        return {k: v for k, v in response.items() if k not in keys_to_omit and v is not None}

    def _generate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        result = LLMResult(generations=[], llm_output={"token_usage": {}})
        if len(prompts) == 0:
            return result

        params = to_model_instance(self.params, GenerateParams)
        params.stop_sequences = stop or params.stop_sequences
        if params.stream:
            if len(prompts) != 1:
                raise GenAiException(ValueError("Streaming works only for a single prompt."))

            generation = GenerationChunk(text="", generation_info={})
            for chunk in self._stream(
                prompt=prompts[0],
                stop=params.stop_sequences,
                run_manager=run_manager,
                **kwargs,
            ):
                self._update_llm_result(current=result, generation_info=chunk.generation_info)
                generation.text += chunk.text
                generation.generation_info.update(self._create_generation_info(chunk.generation_info or {}))

            result.generations.append([generation])
            return result

        model = Model(model=self.model, params=params, credentials=self.credentials)
        for response in model.generate(
            prompts=prompts,
            **kwargs,
        ):
            generation = Generation(
                text=response.generated_text or "",
                generation_info=self._create_full_generation_info(response.dict()),
            )
            logger.info("Output of GENAI call: {}".format(generation.text))
            result.generations.append([generation])
            self._update_llm_result(current=result, generation_info=generation.generation_info)

        return result

    async def _agenerate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self._generate, prompts, stop, run_manager, **kwargs)
        )

    def _stream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Call the IBM GENAI's inference endpoint which then streams the response.
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
            run_manager: Optional callback manager.
        Returns:
            The iterator which yields generation chunks.
        Example:
            .. code-block:: python
                llm = LangChainInterface(model_id="google/flan-ul2", credentials=creds)
                for chunk in llm.stream("What is a molecule?"):
                    print(chunk.text)
        """
        params = to_model_instance(self.params, GenerateParams)
        params.stop_sequences = stop or params.stop_sequences

        model = Model(model=self.model, params=params, credentials=self.credentials)
        for response in model.generate_stream(prompts=[prompt], **kwargs):
            logger.info("Chunk received: {}".format(response.generated_text))
            yield GenerationChunk(
                text=response.generated_text or "",
                generation_info=self._create_full_generation_info(response.dict()),
            )
            if run_manager:
                run_manager.on_llm_new_token(token=response.generated_text, response=response)



import logging

import pytest

from genai import Credentials, Model
from genai.schemas import GenerateParams

logger = logging.getLogger()
logger.addHandler(logging.StreamHandler())


@pytest.mark.unit
class TestLogging:
    def test_no_leaked_logs(self, caplog):
        credentials = Credentials("GENAI_API_KEY")
        params = GenerateParams()
        Model("google/flan-ul2", params=params, credentials=credentials)

        assert len(caplog.records) == 0

    def test_basic_logs(self, caplog):
        caplog.set_level(logging.DEBUG)

        credentials = Credentials("GENAI_API_KEY")
        params = GenerateParams()
        Model("google/flan-ul2", params=params, credentials=credentials)

        assert "google/flan-ul2" in caplog.text



from unittest import TestCase
from unittest.mock import AsyncMock, patch

import pytest

from genai.exceptions import GenAiException
from genai.services import ServiceInterface
from tests.assets.response_helper import SimpleResponse

# API Reference : https://workbench.res.ibm.com/docs/api-reference#generate


@pytest.mark.unit
class TestServiceInterfaceAsync:
    def setup_method(self):
        # mock object for the API call
        self.service = ServiceInterface(service_url="http://SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    # GENERATE ASYNC
    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_generate")
    async def test_generate_async_api_call(self, mock: AsyncMock):
        expected_resp = SimpleResponse.generate(model=self.model, inputs=self.inputs)
        expected = AsyncMock(status_code=200, json=expected_resp)

        mock.return_value = expected
        resp = await self.service.async_generate(model=self.model, inputs=self.inputs)

        assert resp == expected
        assert resp.status_code == 200
        TestCase().assertDictEqual(resp.json, expected_resp)

    @pytest.mark.asyncio
    @patch(
        "genai.services.RequestHandler.async_generate",
        side_effect=Exception("oh no, an exception"),
    )
    async def test_generate_async_with_exception(self, mock):
        with pytest.raises(GenAiException, match="oh no, an exception"):
            await self.service.async_generate(model=self.model, inputs=self.inputs)

    @pytest.mark.asyncio
    async def test_generate_async_with_no_model(self):
        with pytest.raises(TypeError):
            await self.service.async_generate(inputs=self.inputs)

    @pytest.mark.asyncio
    async def test_generate_async_with_no_input(self):
        with pytest.raises(TypeError):
            await self.service.async_generate(model=self.model)

    @pytest.mark.asyncio
    async def test_generate_async_empty_body(self):
        with pytest.raises(TypeError):
            await self.service.async_generate()

    # TOKENIZE  ASYNC
    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_tokenize")
    async def test_tokenize_async_api_call(self, mock: AsyncMock):
        expected_resp = SimpleResponse.tokenize(model=self.model, inputs=self.inputs)
        expected = AsyncMock(status_code=200, json=expected_resp)

        mock.return_value = expected
        resp = await self.service.async_tokenize(model=self.model, inputs=self.inputs)

        assert resp == expected
        assert resp.status_code == 200

    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_tokenize", side_effect=Exception("oh no"))
    async def test_tokenize_async_with_exception(self, mock):
        with pytest.raises(GenAiException, match="oh no"):
            await self.service.async_tokenize(model=self.model, inputs=self.inputs)

    @pytest.mark.asyncio
    async def test_tokenize_async_with_no_model(self):
        with pytest.raises(TypeError):
            await self.service.async_tokenize(inputs=self.inputs)

    @pytest.mark.asyncio
    async def test_tokenize_async_with_no_input(self):
        with pytest.raises(TypeError):
            await self.service.async_tokenize(model=self.model)

    @pytest.mark.asyncio
    async def test_tokenize_async_empty_body(self):
        with pytest.raises(TypeError):
            await self.service.async_tokenize()

    # HISTRORY ASYNC
    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_get")
    async def test_history_async_api_call(self, mock: AsyncMock):
        expected_resp = SimpleResponse.history()
        expected = AsyncMock(status_code=200, json=expected_resp)

        mock.return_value = expected
        resp = await self.service.async_history()

        assert resp == expected
        assert resp.status_code == 200

    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_get", side_effect=Exception("oh no no"))
    async def test_history_async_with_exception(self, mock):
        with pytest.raises(GenAiException, match="oh no no"):
            await self.service.async_history()

    # TOU ASYNC
    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_patch")
    async def test_tou_async_api_call(self, mock: AsyncMock):
        expected_resp = SimpleResponse.terms_of_use()
        expected = AsyncMock(status_code=200, json=expected_resp)

        mock.return_value = expected
        resp = await self.service.async_terms_of_use(True)

        assert resp == expected
        assert resp.status_code == 200

    @pytest.mark.asyncio
    @patch("genai.services.RequestHandler.async_patch", side_effect=Exception("oh no no"))
    async def test_tou_async_with_exception(self, mock):
        with pytest.raises(BaseException, match="oh no no"):
            await self.service.async_terms_of_use(False)



# from unittest.mock import MagicMock, patch

import pytest

from genai.utils.json_utils import json_extract, json_get_all_keys


@pytest.mark.unit
class TestJsonUtils:
    @pytest.fixture
    def json_obj(self):
        return {"key1": "val1", "key2": {"key21": "val21"}, "key3": [{"key31": "val31"}, {"key31": "val32"}]}

    def test_json_extract(self, json_obj):
        a = json_extract(obj=json_obj, key="key1", join=True)
        assert a == "val1"

        a = json_extract(obj=json_obj, key="key31", join=False)
        assert a.sort() == ["val31", "val32"].sort()

        a = json_extract(obj=json_obj, key="key31", join=True)
        assert a == "val31val32"

        a = json_extract(obj=json_obj, key="key4")
        assert a == []

    def test_json_get_all_keys(self, json_obj):
        keys = json_get_all_keys(json_obj)
        assert keys.sort() == ["key1", "key21", "key2", "key3", "key31"].sort()

        keys = json_get_all_keys(json_obj, join=True)
        assert isinstance(keys, str)
        assert keys == "key1key21key31key31"



import os
import pathlib

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.options import Options
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_url = os.getenv("GENAI_API", None)
PATH = pathlib.Path(__file__).parent.resolve()

creds = Credentials(api_key, api_endpoint=api_url)
params = GenerateParams(temperature=0.5)

model = Model("google/flan-ul2", params=params, credentials=creds)


_template = """
{{instruction}}

{{#list}}

Country: {{country}}
Airport: {{airport}}
Capital: {{capital}}

{{/list}}
Country: {{input}}
"""

print("\n------------- Mustaches Prompt Template -------------\n")
pt = PromptPattern.from_watsonx(credentials=creds, name="list-qa-airport-3", template=_template)
print(f"\nPrompt: {pt}")


print("\n------------- Response -------------\n")
options = Options(
    watsonx_template=pt,
    watsonx_data={
        "instruction": "blaahh",
        "list": [
            {"country": "Canada", "capital": "Ottawa", "airport": "YOW"},
            {"country": "Germany", "capital": "Berlin", "airport": "BER"},
            {"country": "USA", "capital": "Washington", "airport": "DCA"},
        ],
    },
)

inputs = ["Spain", "Finland", "Iraq", "India", "Bangladesh"]
for resp in model.generate_async(inputs * 10, options=options, hide_progressbar=True):
    print(f"\nCountry: {resp.input_text} \n{resp.generated_text}")


for resp in model.tokenize_async(inputs, options=options, return_tokens=True):
    print(resp)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../../../src/genai"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "IBM Generative AI Python SDK (Tech Preview)"
copyright = "2023, IBM Research"
author = "IBM Research"

# -- General configuration ---------------------------------ß------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
# html_theme = "alabaster"
html_static_path = ["_static"]
html_title = "IBM Generative AI Python SDK (Tech Preview)"
html_theme_options = {
    "light_css_variables": {
        "font-stack": "'IBM Plex Sans', -apple-system, BlinkMacSystemFont, Segoe UI, Arial, sans-serif",
        "font-stack--monospace": "'IBM Plex Mono', 'SFMono-Regular', Menlo, Consolas, Lucida Console, monospace",
    }
}
html_css_files = ["custom.css"]



import contextlib
import datetime
import logging
import threading
import time
import uuid

try:
    import uvicorn
    from fastapi import APIRouter, FastAPI, Request, status
    from fastapi.responses import JSONResponse
    from starlette.middleware.base import BaseHTTPMiddleware
except ImportError as iex:
    raise ImportError(f"Could not import {iex.name}: Please install ibm-generative-ai[localserver] extension.")


from genai import Credentials
from genai.extensions.localserver.custom_model_interface import CustomModel
from genai.extensions.localserver.schemas import (
    GenerateRequestBody,
    TokenizeRequestBody,
)
from genai.schemas.responses import (
    ErrorResponse,
    GenerateLimits,
    GenerateResponse,
    TokenizeResponse,
)

logger = logging.getLogger(__name__)


class ApiAuthMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, api_key: str = None, insecure: bool = False):
        super().__init__(app)
        self.api_key = api_key
        self.insecure = insecure

    async def dispatch(self, request: Request, call_next):
        auth_header = request.headers["Authorization"]
        logging.debug(f"Incoming Request: Auth: {auth_header}, Endpoint: {request.url}")
        if self.insecure is False:
            if str(self.api_key) not in auth_header:
                logger.warning(f"A client used an incorrect API Key: {auth_header}")
                response_obj = ErrorResponse(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    error="Unauthorized",
                    message="API key not found",
                )
                logger.debug(response_obj.dict())
                return JSONResponse(
                    content=response_obj.dict(),
                    status_code=status.HTTP_401_UNAUTHORIZED,
                )
        logging.debug("Calling next in chain")
        response = await call_next(request)
        logging.debug("Returning response")
        return response


class LocalLLMServer:
    def __init__(
        self,
        models: list[CustomModel],
        port: int = 8080,
        interface: str = "0.0.0.0",
        api_key: str = None,
        insecure_api: bool = False,
    ):
        self.models = {model.model_id: model() for model in models}
        self.port = port
        self.interface = interface
        self.insecure_api = insecure_api

        # Set the API Key
        if api_key is None and insecure_api is False:
            self.api_key = uuid.uuid4()
        elif api_key is None and insecure_api is True:
            self.api_key = "test"

        self.app = FastAPI(
            title="IBM Generative AI Local Model Server",
            debug=False,
            openapi_url=None,
            docs_url=None,
            redoc_url=None,
        )
        self.router = APIRouter()
        self.router.add_api_route("/v1/tokenize", self._route_tokenize, methods=["POST"])
        self.router.add_api_route("/v1/generate", self._route_generate, methods=["POST"])
        self.router.add_api_route("/v1/generate/limits", self._route_generate_limits, methods=["GET"])
        self.app.include_router(self.router)
        self.app.add_middleware(ApiAuthMiddleware, api_key=self.api_key, insecure=insecure_api)

        self.uvicorn_config = uvicorn.Config(self.app, host=interface, port=port, log_level="error", access_log=False)
        self.uvicorn_server = uvicorn.Server(self.uvicorn_config)
        self.endpoint = f"http://{self.interface}:{self.port}"
        logger.debug(f"{__name__}: Models: {list(self.models.keys())}, API: {self.endpoint}, Insecure: {insecure_api}")

    @contextlib.contextmanager
    def run_locally(self):
        thread = threading.Thread(target=self.start_server)
        thread.start()
        try:
            while not self.uvicorn_server.started:
                time.sleep(1e-3)
            yield
        finally:
            self.uvicorn_server.should_exit = True
            thread.join()

    def get_credentials(self):
        return Credentials(api_key=self.api_key, api_endpoint=self.endpoint)

    def start_server(self):
        logger.debug("Starting server background process")

        logger.info(f"Credentials: Endpoint: {self.endpoint} API Key: {self.api_key}")
        logger.info(
            f"In Python Script: credentials = Credentials(api_key={self.api_key}, api_endpoint={self.endpoint})"
        )
        self.uvicorn_server.run()

    def stop_server(self):
        logger.debug("Stopping server background process")
        self.uvicorn_server.shutdown()

    async def _route_tokenize(self, tokenize_request: TokenizeRequestBody):
        logger.info(f"Tokenize called: {tokenize_request}")

        results = [
            self.models[tokenize_request.model_id].tokenize(input_text=input, params=tokenize_request.parameters)
            for input in tokenize_request.inputs
        ]

        created_at = datetime.datetime.now().isoformat()
        response = TokenizeResponse(model_id=tokenize_request.model_id, created_at=created_at, results=results)

        return response

    async def _route_generate(self, generate_request: GenerateRequestBody):
        logger.info(f"Generate Called: Model: {generate_request.model_id}: {generate_request.inputs}")
        results = [
            self.models[generate_request.model_id].generate(input_text=input, params=generate_request.parameters)
            for input in generate_request.inputs
        ]
        created_at = datetime.datetime.now().isoformat()
        response = GenerateResponse(
            id=str(uuid.uuid4()),
            model_id=generate_request.model_id,
            created_at=created_at,
            results=results,
        )
        return response

    async def _route_generate_limits(self):
        logger.info("Generate Limits Called")
        response = GenerateLimits(tokenCapacity=100, tokensUsed=0)
        return response
